/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#include "charging_pulse.h"

#include <list>
#include <set>

#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/shared_ptr.hpp>

#include "hwdesc.h"
#include "hwdesc_attributes.h"
#include "util.h"

namespace dSA
{

typedef std::vector<boost::shared_ptr<Device> > temp_t;

/**
 * @ret only true if lhs strictly smaller then rhs
 * @see: strict weak ordering
 */
bool sortByTemperature(const boost::shared_ptr<const Device> &lhs,
                       const boost::shared_ptr<const Device> &rhs)
{
    assert(rhs && lhs);
    // when using '<=' instead of '<'
    // 1. if lhs == rhs and lhs/rhs == container.begin(); then
    // 2.   std::sort dereferences container.begin()-- -> bang
    return lhs->getLastKnownTemperature() < rhs->getLastKnownTemperature();
}

/* draw n numbers from an urn, without returning */
static std::set<unsigned> sugar_the_cake(unsigned draws, unsigned interval)
{
    assert(charge_slot_count % 4 == 0);
    assert(draws <= interval);

    std::set<unsigned> result;

    if (draws / 2 > 0)
    {
        std::set<unsigned> left, right;
        left = sugar_the_cake(draws / 2, interval / 2);
        right = sugar_the_cake(draws / 2, interval / 2);

        result.insert(left.begin(), left.end());
        BOOST_FOREACH(unsigned draw, right)
        {
            result.insert(draw + interval / 2);
        }
    }

    while (result.size() < draws)
    {
        result.insert(rand() % interval);
    }

    assert(result.size() == draws);
    return result;
}

static std::vector<unsigned> calc_distribution(unsigned samples)
{
    /*
     * fill each slots with x devices, might be 0 if
     * samples < charge_slot_count
     */
    std::vector<unsigned> dist(charge_slot_count, samples / charge_slot_count);

    /*
     * The simplest approach is to start at slot-0 and fill up with each slot
     * with roundup(#devices, #slot) until we run out of devices. But that way
     * we hardly fill the upper slots, unless the number of devices is a
     * multiple of #slot. This is not acceptable, as all slots have to be used
     * equally!
     * Solution: We distribute the remainder devices, randomly among #slot
     */

    int remainder = samples % charge_slot_count;
    if (remainder > 0)
    {
        BOOST_FOREACH(unsigned slot,
                      sugar_the_cake(remainder, charge_slot_count))
        {
            assert(slot < charge_slot_count);
            dist[slot]++;
        }
    }
    return dist;
}


cp_assignments_t assignChargingPulse(const devices_t &devices)
{
    cp_assignments_t result;

    // sort needs random access iterator -> vector
    temp_t sorted_devs;

    // replace with std::copy_if -- requires C++-11
    BOOST_FOREACH(devices_t::value_type elt, devices)
    {
        if (elt->isMainDevice())
        {
            sorted_devs.push_back(elt);
        }
    }

    if (sorted_devs.empty())
    {
        return result;
    }

    std::sort(sorted_devs.begin(), sorted_devs.end(), sortByTemperature);
    std::reverse(sorted_devs.begin(), sorted_devs.end()); // hot devices first

    assert(std::count_if(devices.begin(), devices.end(),
                         boost::bind(&Device::isMainDevice, _1)) ==
           static_cast<int>(sorted_devs.size()));

    std::vector<unsigned> dist = calc_distribution(sorted_devs.size());
    assert(dist.size() == charge_slot_count);

    // dist = [ 2, 3, 3, 3, ... ]
    temp_t::iterator cur = sorted_devs.begin();
    for (unsigned slot = 0; slot < charge_slot_count; slot++)
    {
        for (unsigned j = 0; j < dist[slot]; j++)
        {
            assert(cur != sorted_devs.end());
            result.push_back(std::make_pair(*cur++, slot));
        }
    }

    return result;
}

} // namespace
