/*
    Copyright (c) 2010, 2011, 2012 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>
            Michael Troß, aizo GmbH <michael.tross@aizo.com>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <boost/pointer_cast.hpp>
#include <boost/make_shared.hpp>
#include <errno.h>
#include <string.h>
#include <exception>
#include <iomanip>

#include <digitalSTROM/dsuid.h>
#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>

#include "tasks.h"

#include "common.h"
#include "log.h"
#include "util.h"
#include "exceptions.h"
#include "config_manager.h"
#include "hwdesc.h"
#include "hwstats.h"
#include "storage.h"
#include "device-config.h"
#include "database.h"
#include "ds-assistant.h"
#include "database.h"
#include "util.h"
#include "traffic-decode.h"
#include "gettext_helper.h"
#include "ds_util.h"
#include "comm-channel.h"
#include "vdc-connection.h"
#include "hwdesc_attributes.h"
#include "dsmapi-procedures.h"

#define DSM_API_TIMEOUT_RETRIES 5

#define DSM_TEMP_LOG_INTERVAL   (60*60) // 60 minutes, value in seconds
#define DSM_MAX_TEMP_LOG_INTERVAL (24*60*60) // every 24 hours, value in seconds

namespace dSA
{

dSTaskInfo::dSTaskInfo() : m_name("Generic task"), m_description(""),
                           m_valid(false) {}

void dSTaskInfo::setDescription(std::string description)
{
    m_description = description;
}

const std::string dSTaskInfo::getName() const
{
    return m_name;
}

const std::string dSTaskInfo::getDescription() const
{
    return m_description;
}

bool dSTaskInfo::isValid() const
{
    return m_valid;
}

void dSTaskInfo::invalidate(bool shutdown)
{
    (void)shutdown;
    m_valid = false;
}

dSTaskInfo::~dSTaskInfo() {}

dSTask::dSTask() : dSTaskInfo() {}

dSLongTaskInfo::dSLongTaskInfo() : dSTaskInfo(), m_task_id(INVALID_TASK_ID),
                                    m_parent_task_id(INVALID_TASK_ID),
                                    m_progress(0),
                                    m_status_info(_("Unknown")),
                                    m_task_status(UnknownStatus),
                                    m_started(time(NULL)),
                                    m_finished(0),
                                    m_queued(0),
                                    m_paused(false),
                                    m_pausable(false),
                                    m_task_priority(Normal),
                                    m_persistent(false) {}

dSLongTaskInfo::~dSLongTaskInfo() {}

void dSLongTaskInfo::setTaskID(long long int task_id)
{
    m_task_id = task_id;
}

long long int dSLongTaskInfo::getTaskID() const
{
    return m_task_id;
}

void dSLongTaskInfo::setParentTaskID(long long int task_id)
{
    m_parent_task_id = task_id;
}

long long int dSLongTaskInfo::getParentTaskID() const
{
    return m_parent_task_id;
}

unsigned char dSLongTaskInfo::getProgress() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_progress;
}

std::string dSLongTaskInfo::getProgressInfo() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_progress_info;
}

std::string dSLongTaskInfo::getStatusInfo() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_status_info;
}

std::string dSLongTaskInfo::getParameters() const
{
    return m_parameters;
}

void dSLongTaskInfo::setTaskStatus(task_status status)
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    m_task_status = status;
    switch (status)
    {
        case Running:
            m_status_info = _("Running");
            break;
        case Queued:
            m_status_info = _("Queued");
            break;
        case Cancelled:
            m_status_info = _("Cancelled");
            break;
        case Finished:
            m_status_info = _("Finished");
            break;
        case Suspended:
            m_status_info = _("Suspended");
            break;
        case UnknownStatus:
        default:
            m_status_info = _("Unknown");
            break;
    }
}

task_status dSLongTaskInfo::getTaskStatus() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_task_status;
}

std::string dSLongTaskInfo::getTaskStatusText() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    switch (m_task_status)
    {
        case Queued:
            return "queued";
        case Running:
            return "running";
        case Finished:
            return "finished";
        case Cancelled:
            return "cancelled";
        case Sleeping:
            return "sleeping";
        case Paused:
            return "paused";
        case Suspended:
            return "suspended";
        default:
            break;
    }
    return "unknown";
}

time_t dSLongTaskInfo::getStartedTime() const
{
    return m_started;
}

void dSLongTaskInfo::setStartedTime(time_t t)
{
    m_started = t;
}

time_t dSLongTaskInfo::getFinishedTime() const
{
    return m_finished;
}

void dSLongTaskInfo::setFinishedTime(time_t t)
{
    m_finished = t;
}

time_t dSLongTaskInfo::getQueuedTime() const
{
    return m_queued;
}

void dSLongTaskInfo::setQueuedTime(time_t t)
{
    m_queued = t;
}

bool dSLongTaskInfo::isPaused() const
{
    return m_paused;
}

void dSLongTaskInfo::pause()
{
    m_paused = true;
}

bool dSLongTaskInfo::isPausable() const
{
    return m_pausable;
}

void dSLongTaskInfo::setPriority(task_priority priority)
{
    m_task_priority = priority;
}

task_priority dSLongTaskInfo::getPriority() const
{
    return m_task_priority;
}

bool dSLongTaskInfo::isPersistent() const
{
    return m_persistent;
}

dSLongTask::dSLongTask() : dSLongTaskInfo() {}

dSPersistentTask::dSPersistentTask() : dSLongTask()
{
    m_persistent = true;
}

TaskScheduler::TaskScheduler(boost::shared_ptr<dSTask> task,
                             unsigned int seconds) :
                             TimerSubscriber(seconds, true),
                             m_task(task)
{}

boost::shared_ptr<dSTask> TaskScheduler::getTask()
{
    return m_task;
}

void TaskScheduler::notify()
{
    FILE_LOG(logDEBUG2) <<  "timer notify for task " << m_task->getName();
    dS_Assistant::getInstance()->addTask(m_task);
}



LongTaskScheduler::LongTaskScheduler(boost::shared_ptr<dSLongTask> task,
                                     unsigned int seconds) :
                                     TimerSubscriber(seconds, true),
                                     m_task(task)
{}

boost::shared_ptr<dSLongTask> LongTaskScheduler::getTask()
{
    return m_task;
}

void LongTaskScheduler::notify()
{
    FILE_LOG(logDEBUG2) <<  "timer notify for long task " << m_task->getName();
    if (m_task->isValid())
    {
        dS_Assistant::getInstance()->addLongTask(m_task);
    }
}


GetBusMembersTask::GetBusMembersTask(bool async) : dSTask(), m_async(async)
{
    m_name = _("Get bus members");
    m_description = _("Scan the dS485 bus in order to retrieve information "
                      "about all available devices");
    m_valid = true;
}


void GetBusMembersTask::run(dS_Assistant *dsa)
{
    dsuid_t device_list[DSM_API_MAX_DEVICES];

    if (!(dsa && m_valid))
    {
        return;
    }

    int devices = dsa->dS485()->getBusMembersPerType(device_list,
                                              BUS_MEMBER_GET_DSM | BUS_MEMBER_GET_VDC);
    for (int i = 0; i < devices; ++i)
    {
        boost::shared_ptr<GetDSMInfoTask> task(
                              new GetDSMInfoTask(device_list[i], m_async));
        if (m_async)
        {
            dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
        }
        else
        {
            int retries = 0;
            while (retries < DSM_API_TIMEOUT_RETRIES)
            {
                try
                {
                    task->run(dsa);
                    retries = DSM_API_TIMEOUT_RETRIES;
                }
                catch (Exception &ex)
                {
                    if (ex.getErrorCode() < 0)
                    {
                        retries++;
                    }
                    else
                    {
                        break;
                    }
                }
            } // while
        } // else
    }
}

CheckBusMemberTypeTask::CheckBusMemberTypeTask(dsuid_t dsuid, int flag, bool async) :
                      dSTask(), m_dsuid(dsuid), m_flag(flag), m_async(async)
{
    m_name = _("Check type of bus member");
    m_description = _P(_("Check if bus member %s is a dSM or vDC"),
                                util::str(dsuid).c_str());
    m_valid = true;
}

void CheckBusMemberTypeTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    dsuid_t ownDsuid = dsa->getDSUID();
    if (dsuid_equal(&ownDsuid, &m_dsuid))
    {
        return;
    }

    boost::shared_ptr<dSM> dsm;
    if (m_flag == 0)
    {
        int type = dsa->dS485()->getBusMemberType(m_dsuid);
        if (!BUSMEMBER_IS_DSM(type))
        {
            return;
        }

        try
        {
            dsm = Storage::getInstance()->getDSM(m_dsuid);
        }
        catch (SQLite3Exception &sx)
        {
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                FILE_LOG(logERROR) <<
                        "Unrecoverable SQLite error, terminating dSA";
                g_shutdown_flag = SHUTDOWN_SQLITE;
                return;
            }
            dsm = boost::shared_ptr<dSM>(new dSM(m_dsuid));
            if (type == BusMember_vDC)
            {
                dsm->setVirtual();
            }
        }
        catch (Exception &ex)
        {
            dsm = boost::shared_ptr<dSM>(new dSM(m_dsuid));
            if (type == BusMember_vDC)
            {
                dsm->setVirtual();
            }
        }

        DSUID_LOG(logINFO, m_dsuid) << util::str(m_dsuid) << " joined";
        dsm->setPresent(1);
        boost::shared_ptr<GetDSMInfoTask> task(
                              new GetDSMInfoTask(m_dsuid));
        dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
    else
    {
        try
        {
            dsm = Storage::getInstance()->getDSM(m_dsuid);
        }
        catch (Exception &ex)
        {
            return;
        }

        DSUID_LOG(logINFO, m_dsuid) << util::str(m_dsuid) << " left";
        dsm->setPresent(0);
    }
    Storage::getInstance()->addDSM(dsm);
}

GetDSMInfoTask::GetDSMInfoTask(dsuid_t dsuid, bool async) : dSTask(),
                                                            m_dsuid(dsuid),
                                                            m_async(async)
{
    m_name = _("Request dSM Information");
    m_description = _P(_("Retrieve information for dSM %s"),
                                util::str(dsuid).c_str());
    m_valid = true;
    m_retry = 0;
}

void GetDSMInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint32_t hwv;
    uint32_t armv;
    uint32_t dspv;
    uint16_t apiv;
    dsid_t dsid;
    uint8_t name[21];
    uint32_t cHash, cVersion;
    uint64_t configId, configIdOld;
    int db_old;
    int db_current;
    int type = -1;
    try
    {
        dsa->dS485()->dSMInfo(m_dsuid, &hwv, &armv, &dspv, &apiv, &dsid, name);
        dsa->dS485()->getConfigurationId(m_dsuid, &cHash, &cVersion);
        type = dsa->dS485()->getBusMemberType(m_dsuid);
    }
    catch (Exception& ex)
    {
        if (!m_async)
        {
            throw ex;
        }
        FILE_LOG(logERROR) << "[Task DSMInfo] Error getting "
                "configuration from dSM " << util::str(m_dsuid) <<
                ": " << ex.getMessage();
        if ((ex.getErrorCode() < 0) && (m_retry < DSM_API_TIMEOUT_RETRIES))
        {
            dsa->scheduleTask(boost::shared_ptr<dSTask> (shared_from_this()), 5);
        }
        m_retry ++;
        return;
    }

    // Avoid getting larger values than 2^63-1, they are treated as float
    // when stored into the sqlite table. SQLite uses *signed* integer values:
    //
    // "INTEGER. The value is a signed integer, stored in 1, 2, 3, 4, 6, or 8 bytes
    //  depending on the magnitude of the value."
    //
    configId = (uint64_t) cVersion << 32 | cHash;

    FILE_LOG(logINFO) << "[Task DSMInfo] dSM: " << util::str(m_dsuid) <<
            ", ARM:" << util::hex(armv) <<
            ", DSP:" << util::hex(dspv) <<
            ", HW:" << util::hex(hwv) <<
            ", CFG:" << util::hex(configId);

    boost::shared_ptr<dSM> dsm;
    try
    {
        dsm = Storage::getInstance()->getDSM(m_dsuid);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load dSM: " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            throw sx;
        }
        dsm = boost::shared_ptr<dSM>(new dSM(m_dsuid));
    }
    catch (Exception &ex)
    {
        dsm = boost::shared_ptr<dSM>(new dSM(m_dsuid));
    }
    configIdOld = dsm->getConfigurationId();

    if (type == BusMember_vDC)
    {
        dsm->setVirtual();
    }
    dsm->setPresent(1);
    dsm->setHWVersion(hwv);
    dsm->setARMSwVersion(armv);
    dsm->setDSPSwVersion(dspv);
    dsm->setAPIVersion(apiv);
    dsm->setName(std::string((const char*)name));
    Storage::getInstance()->addDSM(dsm);

    // Synchronously verify meter configuration
    boost::shared_ptr<StartupDSMTask> task(new StartupDSMTask(m_dsuid, false));
    task->run(dsa);

    // The startup task may have restarted the dSM, no use to proceed here now.
    // After the dSM is back and up again we will come back to here anyway.
    dsm = Storage::getInstance()->getDSM(m_dsuid);
    if (! dsm->getPresent())
    {
        FILE_LOG(logDEBUG) << "[Task DSMInfo] dSM " << util::str(m_dsuid) <<
                " not present, stop";
        return;
    }

    // database version 3 introduces new device parameter,
    // so we will have to re-write device information into the db
    // even if the meter configuration has not changed
    Database::getInstance()->getVersionInfo(&db_current, &db_old);

    if ((configId != configIdOld) || (db_old < 3))
    {
        FILE_LOG(logDEBUG) << "New configuration on dSM " << util::str(m_dsuid) <<
                ": " << util::hex(configIdOld) << " -> " << util::hex(configId);

        dsm->setConfigurationId(configId);
        Storage::getInstance()->addDSM(dsm);

        boost::shared_ptr<ScanDSMTask> task(new ScanDSMTask(m_dsuid, m_async));
        if (m_async)
        {
            dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
        }
        else
        {
            int retries = 0;
            while (retries < DSM_API_TIMEOUT_RETRIES)
            {
                try
                {
                    task->run(dsa);
                    retries = DSM_API_TIMEOUT_RETRIES;
                }
                catch (Exception &ex)
                {
                    if (ex.getErrorCode() < 0)
                    {
                        retries++;
                    }
                    else
                    {
                        break;
                    }
                }
            } // while
        }
    }
    else
    {
        // clear deleteflag for every device on this dsm in case dSM
        // configuration has not changed
        Database::getInstance()->setDeviceDeleteFlag(m_dsuid, false);
    }
}

ScanDSMTask::ScanDSMTask(dsuid_t dsuid, bool async) : dSTask(), m_dsuid(dsuid),
                                                    m_async(async)
{
    m_name = _("Scanning dSM for devices");
    m_description = _("Retrieve information about devices attached to the "
                      "dS-Meter");
    m_valid = true;
    m_retry = 0;
}

void ScanDSMTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    boost::shared_ptr<dSM> dsm;
    try
    {
        dsm = Storage::getInstance()->getDSM(m_dsuid);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load dSM " << util::str(m_dsuid) <<
            ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            throw sx;
        }
        return;
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "[Task ScanDSM] dSM with ID " <<
                util::str(m_dsuid) << " not found in database";
        return;
    }

    if (! dsm->getPresent())
    {
        FILE_LOG(logWARNING) << "[Task ScanDSM] dSM with ID " <<
                util::str(m_dsuid) << " is inactive";
        dsm->setConfigurationId(0);     // invalidate config hash
        Storage::getInstance()->addDSM(dsm);
        return;
    }

    uint16_t num;
    try
    {
        dsa->dS485()->getDevicesCountInZone(m_dsuid, 0, &num);
        dsm->setDeviceCount(num);
    }
    catch (Exception& ex)
    {
        FILE_LOG(logERROR) << "[Task ScanDSM] Error getting "
                "configuration from dSM " << util::str(m_dsuid) <<
                ": " << ex.getMessage();
        if (!m_async)
        {
            throw ex;
        }

        if ((ex.getErrorCode() < 0) && (m_retry < DSM_API_TIMEOUT_RETRIES))
        {
            dsa->scheduleTask(boost::shared_ptr<dSTask> (shared_from_this()), 5);
        }
        m_retry ++;
        return;
    }

    // update storage
    Storage::getInstance()->addDSM(dsm);

    // dsm compatibility - ignore dsm's with incompatible old api
    if (dsm->getAPIVersion() < 0x300)
    {
        FILE_LOG(logWARNING) << "Ignoring devices on dSM " << util::str(m_dsuid)
            << ": dsm-api version " << util::hex(dsm->getAPIVersion(), 4)
            << " is incompatible with the system library";
        return;
    }

    // schedule device scanner task
    boost::shared_ptr<GetAllDevicesTask> task(new GetAllDevicesTask(m_dsuid));
    if (m_async)
    {
        dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
    else
    {
        task->run(dsa);
    }
}
StartupDSMTask::StartupDSMTask(dsuid_t dsuid, bool async) : dSTask(),
                                                  m_dsuid(dsuid), m_async(async)
{
    m_name = _("dSM startup handler");
    m_description = _("Check dSM configuration on startup");
    m_valid = true;
}

void StartupDSMTask::run(dS_Assistant *dsa)
{
    bool restartRequired = false;
    boost::shared_ptr<dSM> dsmconfig;

    try
    {
        dsmconfig = Storage::getInstance()->getDSM(m_dsuid);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load dSM " << util::str(m_dsuid) <<
            ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            throw sx;
        }
        return;
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "[Task StartupDSM] dSM with ID " <<
                util::str(m_dsuid) << " not found in database";
        return;
    }

    if (!(dsa && m_valid))
    {
        return;
    }

    FILE_LOG(logDEBUG) << "[Task StartupDSM] dSM " << util::str(m_dsuid);

    for (int retry = 0; retry < 3; retry ++ )
    {
        // 0. Read dSM State
        try
        {
            uint8_t Hardware, DSP, Temperature, Config, DS485, Line;
            dsa->dS485()->dSMState(m_dsuid, &Hardware, &DSP, &Temperature, &Config, &DS485, &Line);
            DSUID_LOG(logINFO, m_dsuid) << "[Task StartupDSM] "
                    "State Hardware:" << util::str(Hardware) << ", "
                    "CommunicationDSP:" << util::str(DSP) << ", "
                    "Temperature:" << util::str(Temperature) << ", "
                    "Config:" << util::str(Config) << ", "
                    "DS485:" << util::str(DS485) << ", "
                    "Line:" << util::str(Line);

            Database::getInstance()->updateDSMState(m_dsuid,
                    Hardware, DSP, Temperature, Config, DS485, Line);
        }
        catch (Exception &ex)
        {
            DSUID_LOG(logERROR, m_dsuid) << "[Task CheckDSM] "
                    "Failed to query dsm status: " << ex.getMessage();
            throw ex;
        }

        // 1. Verify apartment id
        try
        {
            uint8_t aid;
            int dbid;
            ConfigManager::getInstance()->getParameter(
                    "user.digitalstrom.apartmentid", dbid);
            dsa->dS485()->getApartmentId(m_dsuid, &aid);
            if (aid != dbid)
            {
                dsa->dS485()->setApartmentId(m_dsuid, dbid);
                restartRequired = true;
                DSUID_LOG(logINFO, m_dsuid) << "[Task StartupDSM] ApartmentId "
                "reconfiguration from " << util::str(aid) <<
                " to " << util::str(dbid);
            }
            dsmconfig->setApartmentId(aid);
        }
        catch(Exception ex)
        {
            DSUID_LOG(logERROR, m_dsuid) << "[Task StartupDSM] "
                    "Reconfigure ApartmentId: " << ex.getMessage();
            continue;
        }

        // 2. Verify meter settings, get default modes from db
        try
        {
            uint8_t flags, newflags;
            int pmode;
            bool led_on;
            bool progmode;
            ConfigManager::getInstance()->getParameter(
                    "user.digitalstrom.meter.led", led_on);
            ConfigManager::getInstance()->getParameter(
                    "digitalstrom.meter.pingmode", pmode);
            ConfigManager::getInstance()->getParameter(
                    "user.digitalstrom.programmingmode", progmode);
            dsa->dS485()->getPropertyFlags(m_dsuid, &flags);
            newflags = flags;

            uint8_t fvalue = led_on ? 1 : 0;
            if ((flags & 1) != fvalue)
            {
                newflags &= ~1;
                newflags |= fvalue;
                restartRequired = true;
            }
            if (((flags >> 1) & 3) != pmode)
            {
                newflags &= ~(3 << 1);
                newflags |= (pmode << 1);
                restartRequired = true;
                DSUID_LOG(logDEBUG, m_dsuid) << "[Task StartupDSM] Ping Mode "
                        "reconfiguration " << util::str((flags >> 1) & 3) <<
                        " => " << util::str(pmode);
            }
            fvalue = progmode ? 0 : 8;
            if ((flags & 8) != fvalue)
            {
                newflags &= ~8;
                newflags |= fvalue;
            }

            if (newflags != flags)
            {
                dsa->dS485()->setPropertyFlags(m_dsuid, newflags);
                DSUID_LOG(logINFO, m_dsuid) << "[Task StartupDSM] "
                        "Reconfigure dSM Flags " <<
                        util::hex(flags, 2) << " => " << util::hex(newflags, 2);
            }
        }
        catch(Exception ex)
        {
            DSUID_LOG(logERROR, m_dsuid) << "[Task StartupDSM] "
                    "Reconfigure dSM Flags: " << ex.getMessage();
            continue;
        }

        // 3. Verify temperature limits, get defaults from db (90/150 instead of 70/90)
        try
        {
            uint8_t wt, et;
            int warnlevel;
            int errorlevel;
            ConfigManager::getInstance()->getParameter("digitalstrom.meter.twarning", warnlevel);
            ConfigManager::getInstance()->getParameter("digitalstrom.meter.tfailure", errorlevel);
            dsa->dS485()->getPropertyTemperatureLimit(m_dsuid, &wt, &et);
            if (wt != warnlevel || et != errorlevel)
            {
                dsa->dS485()->setPropertyTemperatureLimit(m_dsuid, warnlevel, errorlevel);
                restartRequired = true;
                DSUID_LOG(logINFO, m_dsuid) << "[Task StartupDSM] Temperature level "
                "reconfiguration from " << util::str(wt) << "/" << util::str(et) <<
                " to " << util::str(warnlevel) << "/" << util::str(errorlevel);
            }
        }
        catch(Exception ex)
        {
            DSUID_LOG(logERROR, m_dsuid) << "[Task StartupDSM] "
                    "Reconfigure Temperature Limits: " << ex.getMessage();
            continue;
        }

        // 4. Check: Communication Log
        dsa->dS485()->setCommunicationLogging(m_dsuid, 0xff);

        // Restart dSM if required, reset present flags
        try
        {
            boost::shared_ptr<dSM> dsm;
            dsm = Storage::getInstance()->getDSM(m_dsuid);
            if (!dsm->isVirtual() && restartRequired)
            {
                dsa->dS485()->restartMeter(m_dsuid);
                dsmconfig->setPresent(0);
                DSUID_LOG(logINFO, m_dsuid) << "[Task StartupDSM] "
                    "Restart dSM : " + util::str(m_dsuid);
            }
        }
        catch(Exception ex)
        {
            DSUID_LOG(logERROR, m_dsuid) << "[Task StartupDSM] "
                    "Restart dSM failed: " << ex.getMessage();
            continue;
        }

        break;
    }

    Storage::getInstance()->addDSM(dsmconfig);
}

SetDSMApartmentIdTask::SetDSMApartmentIdTask(dsuid_t dsuid, int apartment_id,
        bool async) :
        dSTask(), m_dsuid(dsuid), m_apartment_id(apartment_id), m_async(async)
{
    m_name = _("Set apartment id handler");
    m_description = _("Set dSM apartment id and restart");
    m_valid = true;
}

void SetDSMApartmentIdTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    boost::shared_ptr<std::list<dsuid_t> > dsms;
    boost::shared_ptr<std::list<dsuid_t> > dsm_list(new std::list<dsuid_t>);
    if (dsuid_is_null(&m_dsuid))
    {
        dsms = Storage::getInstance()->getDSMList();
    }
    else
    {
        dsms = dsm_list;
        dsms->push_back(m_dsuid);
    }

    // loop over all dSMeter
    for (std::list<dsuid_t>::iterator itm = dsms->begin();
         m_valid && (itm != dsms->end());
         itm++)
    {
        boost::shared_ptr<dSM> dsm;
        uint8_t aid;

        try
        {
            dsm = Storage::getInstance()->getDSM(*itm);
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << "Could not load dSM " << util::str(*itm) <<
                ": " << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                throw sx;
            }
            continue;
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "[Task SetApartmentId] dSM with ID " <<
                    util::str(*itm) << " not found in database";
            continue;
        }

        dsa->dS485()->getApartmentId(*itm, &aid);

        DSUID_LOG(logDEBUG2, *itm) << "[Task SetApartmentId] "
                "Database = " << util::str(m_apartment_id) <<
                ", Datamodel = " << util::str(dsm->getApartmentId()) <<
                ", Device = " << util::str(aid);

        if ((uint8_t) m_apartment_id != aid)
        {
            if (dsm->getApartmentId() >= 64)
            {
                FILE_LOG(logERROR) <<
                    "[Task SetApartmentId] Id mismatch on dsm " <<
                    util::str(*itm) <<
                    ": Database value " << util::str(m_apartment_id) <<
                    ", device value " << util::str(aid) <<
                    ", repeatedly failed to set id #" <<
                    util::str(dsm->getApartmentId() - 64);

                dsm->setApartmentId(dsm->getApartmentId() + 1);

                // TODO: Device Error Management
                // Populate misconfiguration error to dsm status monitor
                // CheckDSMTask
            }

            for (int retry = 0; retry < DSM_API_TIMEOUT_RETRIES; retry ++)
            {
                DSUID_LOG(logINFO, *itm) << "[Task SetApartmentId] Id "
                "reconfiguration from " << util::str(aid) <<
                " to " << util::str(m_apartment_id);

                aid = (uint8_t) m_apartment_id;
                try
                {
                    dsa->dS485()->setApartmentId(*itm, aid);
                }
                catch (Exception &ex)
                {
                    DSUID_LOG(logERROR, *itm) << "[Task SetApartmentId] "
                    "reconfiguration failed: " << ex.getMessage();
                    continue;
                }

                try
                {
                    if (! dsm->isVirtual())
                    {
                        dsa->dS485()->restartMeter(*itm);
                        DSUID_LOG(logINFO, m_dsuid) << "[Task SetApartmentId] "
                            "Restart dSM : " + util::str(*itm);
                    }
                }
                catch(Exception ex)
                {
                    DSUID_LOG(logERROR, *itm) << "[Task SetApartmentId] "
                    "restart failed: " << ex.getMessage();
                    continue;
                }
                break;
            }

            if (dsm->getApartmentId() >= 64)
            {
                dsm->setApartmentId(dsm->getApartmentId() + 1);
            }
            else {
                dsm->setApartmentId(64);
            }
            Storage::getInstance()->addDSM(dsm);
        }
    }
}

CheckDSMTask::CheckDSMTask(dsuid_t dsuid, bool async) : dSTask(), m_dsuid(dsuid),
        m_async(async)
{
    m_name = _("dSM health check handler");
    m_description = _("Check dSM state and configuration");
    m_valid = true;
}

void CheckDSMTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    // 1. Query dSMState() and set global health status accordingly
    {
        boost::shared_ptr<dSM> dsm;
        try
        {
            dsm = Storage::getInstance()->getDSM(m_dsuid);
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << "Could not load dSM " << util::str(m_dsuid) <<
                ": " << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                throw sx;
            }
            return;
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "[Task CheckDSM] dSM with ID " <<
                    util::str(m_dsuid) + " not found in database";
            return;
        }

        bool restartRequired = false;
        try
        {
            uint8_t flags, newflags;
            bool led_on;
            bool progmode;
            bool roomdimmmode;
            ConfigManager::getInstance()->getParameter(
                    "user.digitalstrom.meter.led", led_on);
            ConfigManager::getInstance()->getParameter(
                    "user.digitalstrom.programmingmode", progmode);
            ConfigManager::getInstance()->getParameter(
                    "user.digitalstrom.roomdimmingmode", roomdimmmode);
            dsa->dS485()->getPropertyFlags(m_dsuid, &flags);
            newflags = flags;

            uint8_t fvalue = led_on ? 1 : 0;
            if ((flags & 1) != fvalue)
            {
                newflags &= ~1;
                newflags |= fvalue;
                restartRequired = true;
            }

            fvalue = progmode ? 0 : 8;
            if ((flags & 8) != fvalue)
            {
                newflags &= ~8;
                newflags |= fvalue;
            }

            fvalue = roomdimmmode ? 0 : 0x80;
            if ((flags & 0x80) != fvalue)
            {
                newflags &= ~0x80;
                newflags |= fvalue;
            }

            if (newflags != flags)
            {
                dsa->dS485()->setPropertyFlags(m_dsuid, newflags);
                DSUID_LOG(logINFO, m_dsuid) << "[Task CheckDSM] "
                        "Reconfigure dSM Flags " <<
                        util::hex(flags, 2) << " => " << util::hex(newflags, 2);
            }
        }
        catch(Exception ex)
        {
            DSUID_LOG(logERROR, m_dsuid) << "[Task CheckDSM] "
                    "Reconfigure dSM Flags: " << ex.getMessage();
        }

        // Restart dSM if required, reset present flags
        try
        {
            if (!dsm->isVirtual() && restartRequired)
            {
                dsa->dS485()->restartMeter(m_dsuid);
                dsm->setPresent(0);
                DSUID_LOG(logINFO, m_dsuid) << "[Task StartupDSM] "
                    "Restart dSM : " + util::str(m_dsuid);
            }
        }
        catch(Exception ex)
        {
            DSUID_LOG(logERROR, m_dsuid) << "[Task StartupDSM] "
                    "Restart dSM failed: " << ex.getMessage();
        }

        // TODO: status delivery and further processing?


    }
}

CheckDSMThread::CheckDSMThread(int interval) :
        dSLongTask(), m_interval(interval)
{
    m_name = _("Periodic meter monitor");
    m_description = _("Periodically check dSM state and configuration");
    m_valid = true;
    m_pausable = true;
    m_state = 0;
    m_index = 0;
}

task_type CheckDSMThread::getTaskType() const
{
    return UnknownTask;
}

void CheckDSMThread::run(dS_Assistant *dsa)
{
    bool resetApartmentId = false;
    int newApartmentId = 0;

    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    // Automatic ApartmentId Reconfiguration
    {
        int apartmentid, busmembercount;
        dsuid_t device_list[DSM_API_MAX_DEVICES];
        try
        {
            ConfigManager::getInstance()->getParameter(
                    "user.digitalstrom.apartmentid", apartmentid);
            busmembercount = dS_Assistant::getInstance()->dS485()->getBusMembersPerType(
                    device_list, BUS_MEMBER_GET_DSM | BUS_MEMBER_GET_DS485D);
            if (apartmentid + busmembercount >= 63)
            {
                newApartmentId = 63 - busmembercount;
                newApartmentId &= 0x38;  // 11 1000
                ConfigManager::getInstance()->setParameter(
                        "user.digitalstrom.apartmentid", newApartmentId);
                FILE_LOG(logWARNING) << "[Thread CheckDSM] "
                        "ApartmentId reconfiguration, new base address " <<
                        util::str(newApartmentId);
                resetApartmentId = true;
            }
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "[Thread CheckDSM] "
                    "Failed to query apartment id settings: "
                    << ex.getMessage();
        }
    }

    while (m_valid)
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Paused;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case 0:
            {
                m_dsms = Storage::getInstance()->getDSMList();
                m_index = 0;
                m_state ++;
            }
            break;

            case 1:
            {
                if (m_dsms->empty())
                {
                    m_state ++;
                }
                else
                {
                    dsuid_t dsmdsuid = m_dsms->front();
                    m_dsms->pop_front();
                    m_index ++;

                    boost::shared_ptr<dSM> dsm;
                    try
                    {
                        dsm = Storage::getInstance()->getDSM(dsmdsuid);
                    }
                    catch (SQLite3Exception &sx)
                    {
                        FILE_LOG(logERROR) << "Could not load dSM " <<
                            util::str(dsmdsuid) << ": " << sx.getMessage();
                        if (SQLite3::isFatal(sx.getErrorCode()))
                        {
                            throw sx;
                        }
                    }
                    catch (Exception &ex)
                    {
                    }

                    if (dsm.get() && dsm->getPresent())
                    {
                        boost::shared_ptr<CheckDSMTask>
                            task(new CheckDSMTask(dsmdsuid, false));
                        dsa->addTask(task);

                        if (resetApartmentId)
                        {
                            boost::shared_ptr<SetDSMApartmentIdTask>
                                task2(new SetDSMApartmentIdTask(dsmdsuid,
                                        newApartmentId, false));
                            dsa->addTask(task2);
                        }
                    }
                }
            }
            break;

            default:
            {
                m_status_mutex.lock();
                m_task_status = Finished;
                m_status_mutex.unlock();
                if (m_interval > 0)
                {
                    boost::shared_ptr<CheckDSMThread> task(
                            new CheckDSMThread(m_interval));
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (task), m_interval);
                }
                return;
            }
            break;

        }
    }
}


GetActiveDevicesInZoneTask::GetActiveDevicesInZoneTask(dsuid_t dsuid,
                                                       uint16_t zone_id) :
                                                       dSTask(), m_dsuid(dsuid),
                                                       m_zone_id(zone_id)
{
    m_name = _("Get active devices from zone");
    m_description = _("Retrieve information about active devices in a given "
                      "zone on a dS-Meter");
    m_valid = true;
}

void GetActiveDevicesInZoneTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    if (! ds_util::isDSMCompatible(m_dsuid))
    {
        FILE_LOG(logINFO) << "[Task GetActiveDevices] dSM with ID " <<
                util::str(m_dsuid) << " is not compatible.";
        return;
    }

    uint16_t count;

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t name[21];
    dsuid_t dsuid;
    uint8_t lt_mode;

    dsa->dS485()->getActiveDevicesCountInZone(m_dsuid, m_zone_id, &count);
    for (uint16_t i = 0; i < count; i++)
    {
        dsa->dS485()->getActiveDeviceInfoInZone(m_dsuid, m_zone_id, i,
                                       &device_id,
                                       &vendor_id, &product_id, &function_id,
                                       &version, &zone_id, &state,
                                       &locked, &output_mode, NULL,
                                       name, &dsuid, &lt_mode);

        boost::shared_ptr<Device> dev(new Device(dsuid, m_dsuid));
        dev->setId(device_id);
        dev->setProductId(product_id);
        dev->setFunctionId(function_id);
        dev->setVersion(version);
        dev->setZoneId(zone_id);
        dev->setState(state);
        dev->setLocked(locked);
        dev->setOutputMode(output_mode);
        dev->setName((const char*)name);
        dev->setLTMode(lt_mode);
        Storage::getInstance()->addDevice(dev);

        FILE_LOG(logINFO) << "Added " << (state ? "active" : "inactive")
                          << " device with dSUID " << util::str(dev->getDSUID())
                          << " and address " << util::str(device_id)
                          << " on dSM " << util::str(dev->getDSMdSUID());
    }
}

GetAllDevicesTask::GetAllDevicesTask(dsuid_t dsuid) :
        dSTask(), m_dsuid(dsuid)
{
    m_name = _("Get all devices on a circuit");
    m_description = _("Retrieve information about all devices on a given circuit");
    m_valid = true;
    m_retry = 0;
}

void GetAllDevicesTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t count;

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t name[21];
    dsuid_t dsuid;
    uint8_t lt_mode;

    boost::shared_ptr<dSM> dsm;
    try
    {
        dsm = Storage::getInstance()->getDSM(m_dsuid);
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "[Task GetAllDevicesTask] dSM with ID " <<
                util::str(m_dsuid) << " not found in database";
        return;
    }

    if (! dsm->getPresent())
    {
        FILE_LOG(logWARNING) << "[Task GetAllDevicesTask] dSM with ID " <<
                util::str(m_dsuid) << " is inactive";
        dsm->setConfigurationId(0);     // invalidate config hash
        Storage::getInstance()->addDSM(dsm);
        return;
    }

    if (! dsm->isAPIVersionCompatible())
    {
        FILE_LOG(logINFO) << "[Task GetAllDevicesTask] dSM with ID " <<
                util::str(m_dsuid) << " is not compatible." <<
                "Not reading out device info.";
        return;
    }

    try
    {
        dsa->dS485()->getDevicesCountInZone(m_dsuid, 0, &count);
    }
    catch (Exception& ex)
    {
        FILE_LOG(logERROR) << "[Task GetAllDevices] Error getting "
                "configuration from dSM " << util::str(m_dsuid) <<
                ": " << ex.getMessage();
        if ((ex.getErrorCode() < 0) && (m_retry < DSM_API_TIMEOUT_RETRIES))
        {
            dsa->scheduleTask(boost::shared_ptr<dSTask> (shared_from_this()), 10);
        }
        m_retry ++;
        return;
    }

    for (uint16_t i = 0; i < count; i++)
    {
        try
        {
            dsa->dS485()->getDeviceInfoByIndex(m_dsuid, 0, i,
                    &device_id, &vendor_id, &product_id, &function_id,
                    &version, &zone_id, &state,
                    &locked, &output_mode, NULL,
                    name, &dsuid, &lt_mode);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR) << "[Task GetAllDevices] Error getting "
                    "device configuration from dSM " << util::str(m_dsuid) <<
                    ": " << ex.getMessage();
            if ((ex.getErrorCode() < 0) && (m_retry < DSM_API_TIMEOUT_RETRIES))
            {
                dsa->scheduleTask(boost::shared_ptr<dSTask> (shared_from_this()), 10);
                m_retry ++;
                return;
            }
            continue;
        }

        boost::shared_ptr<Device> dev;

        if (dsm->isVirtual())
        {
            dev = boost::shared_ptr<VirtualDevice>(new VirtualDevice(dsuid, m_dsuid));
            try
            {
                boost::shared_ptr<VirtualDevice> virtualDev =
                    boost::dynamic_pointer_cast<VirtualDevice>(dev);
                std::string hwid = VdcHelper::getHardwareGuid(m_dsuid, dsuid);
                virtualDev->setHardwareGuid(hwid);
                std::string hwinfo = VdcHelper::getHardwareInfo(m_dsuid, dsuid);
                virtualDev->setHardwareInfoString(hwinfo);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << "[Task GetAllDevices] Error getting "
                "hardware guid or capabilities from virtual device on vDC "
                << util::str(m_dsuid) << ": " << ex.getMessage();
            }
        }
        else
        {
            dev = boost::shared_ptr<Device>(new Device(dsuid, m_dsuid));
        }
        dev->setId(device_id);
        dev->setProductId(product_id);
        dev->setFunctionId(function_id);
        dev->setVersion(version);
        dev->setZoneId(zone_id);
        dev->setState(state);
        dev->setLocked(locked);
        dev->setOutputMode(output_mode);
        dev->setName((const char*)name);
        dev->setLTMode(lt_mode);

        //TODO: set zone index

        Storage::getInstance()->addDevice(dev);

        FILE_LOG(logINFO) << "Added " << (state ? "active" : "inactive")  <<
                " device with dSUID " << util::str(dev->getDSUID()) <<
                " and address " << util::str(device_id) <<
                " on dSM " << util::str(dev->getDSMdSUID());

        boost::shared_ptr<CheckDeviceRegistrationTask> tcheck(
                new CheckDeviceRegistrationTask(dsuid));
        dsa->addTask(boost::dynamic_pointer_cast<dSTask>(tcheck));
    }
}

GetDSMZonesTask::GetDSMZonesTask(dsuid_t dsuid) : dSTask(), m_dsuid(dsuid)
{
    m_name = _("Get number of zones on dSM");
    m_description = _P(_("Retrieve number of zones for dSM %s"),
                                util::str(dsuid).c_str());
    m_valid = true;
}

void GetDSMZonesTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint8_t count;
    dsa->dS485()->getZoneCount(m_dsuid, &count);
    for (uint8_t i = 0; i < count; i++)
    {
        boost::shared_ptr<GetDSMZoneInfoTask> task(
                                            new GetDSMZoneInfoTask(m_dsuid, i));
        dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
}

GetDSMZoneInfoTask::GetDSMZoneInfoTask(dsuid_t dsuid, uint8_t zone_index) :
                                       dSTask(), m_dsuid(dsuid),
                                       m_zone_index(zone_index)
{
    m_name = _("Get zone information");
    m_description = _P(_("Retrieve information for zone with index %d"
                                  "on dSM %s"), zone_index,
                                util::str(dsuid).c_str());
    m_valid = true;
}

void GetDSMZoneInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t zone_id;
    uint8_t vzone_id;
    uint8_t num_grp;
    uint8_t name[21];

    dsa->dS485()->getZoneInfo(m_dsuid, m_zone_index, &zone_id, &vzone_id,
                              &num_grp, name);
    FILE_LOG(logINFO) << "Retrieved zone info for dsm: "
        << util::str(m_dsuid) << ", zoneid: " << (int)zone_id << " vzid: "
        << (int)vzone_id << " groups:  " << (int)num_grp << "name: " << name;
}

GetDeviceInfoTask::GetDeviceInfoTask(dsuid_t dsm_dsuid, uint16_t device_id) :
                                     dSTask(), m_dsuid(dsm_dsuid),
                                     m_device_id(device_id)
{
    m_name = _("Get device information");
    m_description = _P(_("Retrieve device information for device "
                                  "on dSM %s"), util::str(dsm_dsuid).c_str());
    m_valid = true;
}

void GetDeviceInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    if (! ds_util::isDSMCompatible(m_dsuid))
    {
        FILE_LOG(logINFO) << "dSM with ID " << util::str(m_dsuid) <<
                " is not compatible. Not reading out device " <<
                util::str(m_device_id);
        return;
    }

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t name[21];
    dsuid_t dsuid;
    uint8_t lt_mode;

    dsa->dS485()->getDeviceInfoById(m_dsuid, m_device_id,
                       &device_id, &vendor_id, &product_id, &function_id,
                       &version, &zone_id, &state, &locked,
                       &output_mode, NULL, name, &dsuid, &lt_mode);

    boost::shared_ptr<Device> dev;

    try
    {
        dev = Storage::getInstance()->getDevice(dsuid, m_dsuid);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "Could not load device " <<
            util::str(dsuid) << ": " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            throw sx;
        }
        dev = boost::shared_ptr<Device>(new Device(dsuid, m_dsuid));
    }
    catch (Exception &ex)
    {
        dev = boost::shared_ptr<Device>(new Device(dsuid, m_dsuid));
    }

    if (dev->getState() == 0 && state > 0)
    {
        DSUID_LOG(logINFO, m_dsuid) << "active";
    }

    dev->setId(device_id);
    dev->setProductId(product_id);
    dev->setFunctionId(function_id);
    dev->setVersion(version);
    dev->setZoneId(zone_id);
    dev->setState(state);
    dev->setLocked(locked);
    dev->setOutputMode(output_mode);
    dev->setName((const char*)name);
    dev->setLTMode(lt_mode);
    Storage::getInstance()->addDevice(dev);

    FILE_LOG(logINFO) << "Updated device info with dSUID "
                      << util::str(dev->getDSUID()) <<  " on dSM "
                      << util::str(dev->getDSMdSUID());

    boost::shared_ptr<CheckDeviceRegistrationTask> tcheck(
            new CheckDeviceRegistrationTask(m_dsuid));
    dsa->addTask(boost::dynamic_pointer_cast<dSTask>(tcheck));
}

CheckDeviceRegistrationTask::CheckDeviceRegistrationTask(dsuid_t dsuid) :
        dSTask(), m_dsuid(dsuid)
{
    m_name = "Check device registration status";
    m_description = "Verify device " + util::str(dsuid) + "is active on a single dSM only";
    m_valid = true;
}

void CheckDeviceRegistrationTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    try
    {
        boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                 devlist = Storage::getInstance()->getDeviceList(m_dsuid, true);
        if (devlist->size() > 1)
        {
            FILE_LOG(logINFO) << "Detected multiple registrations for "
                    "dSUID " << util::str(m_dsuid);

            for (std::list<boost::shared_ptr<Device> >::iterator it =
                    devlist->begin();
                    it != devlist->end();
                    it++)
            {
                boost::shared_ptr<TestOneDevicePresenceTask> task(
                   new TestOneDevicePresenceTask(m_dsuid, (*it)->getDSMdSUID()));
                dsa->addLongTask(task);
            }
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logWARNING) << "Failed to check registration status of "
                "dSUID " << util::str(m_dsuid) << ": " << ex.getMessage();
    }
}

TestTranmissionQualityTask::TestTranmissionQualityTask(int checksPerDevice,
                                dsuid_t dsm_dsuid) : dSLongTask(),
                                m_checksPerDevice(checksPerDevice),
                                m_dsm_dsuid(dsm_dsuid),
                                m_result_down(0), m_result_up(0),
                                m_result_err(0), m_dev_count(0),
                                m_dev_total(0), m_check_loop(0)
{
    m_name = _("Test transmission quality of devices");
    m_parameters = "dsm=" + util::str(dsm_dsuid) + '&' +
                   "checksPerDevice=" + util::str(checksPerDevice);

    if (dsuid_is_null(&dsm_dsuid))
    {
        m_description = _P(_("Test transmission quality for all "
                                    "devices, using %d checks per device"),
                                    checksPerDevice);
    }
    else
    {
        m_description = _P(_("Test transmission quality for all "
                                      "devices on dSM %s, using %d checks "
                                      "per device"),
                                      util::str(dsm_dsuid).c_str(),
                                      checksPerDevice);
    }

    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
}

TestTranmissionQualityTask::~TestTranmissionQualityTask()
{
    try
    {
        dSA::dS_Assistant::getInstance()->dS485()->removeListener(this);
    }
    catch(Exception& e)
    {
    }
}

task_type TestTranmissionQualityTask::getTaskType() const
{
    return TestTransmissionQuality;
}

void TestTranmissionQualityTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {
        default:
            break;

        case DsmApiDeviceTestTransmissionQuality:
        {
            boost::shared_ptr<DeviceTestQualityData> data =
                boost::dynamic_pointer_cast<DeviceTestQualityData> (busdata);
            m_result_up = data->getUpstream();
            m_result_down = data->getDownstream();
            m_result_err = 0;
        }
        break;
    }
}

void TestTranmissionQualityTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    if (kStateInit == m_state)
    {
        if (dsuid_is_null(&m_dsm_dsuid))
        {
            m_dev_list = Storage::getInstance()->getDevices(true);
        }
        else
        {
            m_dev_list = Storage::getInstance()->getDevices(m_dsm_dsuid, true);
        }
        m_dev_count = 0;
        m_dev_current = m_dev_list->begin();
        m_dev_total = m_dev_list->size() * m_checksPerDevice;
        m_check_loop = 0;
        m_progress = 0;

        dsa->dS485()->addListener(this);
        this->select(DsmApiDeviceTestTransmissionQuality, true);

        m_state = kStateSetup;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    if (0 == m_dev_total)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    while (m_valid && (m_dev_count <= m_dev_total) && (m_state != kStateFinished))
    {

        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                m_result_down = TRANSMISSION_QUALITY_DOWNSTREAM_MAX;
                m_result_up = 0;
                m_result_err = ERROR_TIMEOUT;

                boost::shared_ptr<Device> device = *m_dev_current;
                m_dev_count ++;

                m_status_mutex.lock();
                m_status_info = _P(_("Requesting transmission quality"
                            " from device %s"),
                            util::str(device->getDSUID()).c_str());
                if (m_checksPerDevice > 1)
                {
                    m_status_info = _P(
                            _("Requesting transmission quality from device %s"
                              ", test %d"),
                            util::str(device->getDSUID()).c_str(),
                            m_check_loop + 1);
                }
                m_status_mutex.unlock();

                try
                {
                    dsa->dS485()->reqDeviceTransmissionQuality(device->getDSMdSUID(), device->getId());
                    m_state = kStateRequest;
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (shared_from_this()),
                            2);
                    return;
                }
                catch (Exception &ex)
                {
                    m_status_mutex.lock();
                    m_status_info = _P(_("Failed to get transmission "
                        "quality for device %s: %s"),
                            util::str(device->getDSUID()).c_str(),
                            ex.getMessage().c_str());
                    m_result_err = ex.getErrorCode();
                    m_status_mutex.unlock();

                    FILE_LOG(logINFO) << "[Task Transmission Quality] Device " <<
                            util::str(device->getDSUID()) <<
                            ", error " << util::str(m_result_err) << " = "
                            << ex.getMessage();
                    m_state = kStateResponse;
                }
            }
            break;

            case kStateRequest:
            {
                m_state = kStateResponse;
                if (m_result_err == ERROR_TIMEOUT)
                {
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (shared_from_this()),
                            2);
                    return;
                }
            }
            break;

            case kStateResponse:
            {
                m_state = kStateNext;
                boost::shared_ptr<Device> device = *m_dev_current;
                Database::getInstance()->saveTransmissionQuality(m_task_id,
                        util::str(device->getDSUID()),
                        m_result_up, m_result_down, m_result_err);

                m_status_mutex.lock();
                m_progress = (unsigned char)(m_dev_count * 100 / m_dev_total);
                m_status_mutex.unlock();

                FILE_LOG(logDEBUG1) << "[Task Transmission Quality] "
                        "Progress " << util::str(m_progress) <<
                        ", Device " << util::str(device->getDSUID()) <<
                        ", down " << util::str(m_result_down) <<
                        ", up " << util::str(m_result_up);
            }
            break;

            case kStateNext:
            {
                m_state = kStateSetup;
                m_dev_current ++;
                if (m_dev_current == m_dev_list->end())
                {
                    m_check_loop++;
                    if (m_check_loop < m_checksPerDevice)
                    {
                        m_dev_current = m_dev_list->begin();
                    }
                    else
                    {
                        m_state = kStateFinished;
                    }
                }
            }
            break;

            case kStateFinished:
            {
                m_dev_count = m_dev_total;
                FILE_LOG(logWARNING) << "[Task Transmission Quality] Force task end";
            }
            break;

            default:
                m_state = kStateFinished;
                break;

        } // switch
    }

    FILE_LOG(logINFO) << "[Task Transmission Quality] Ready, sent " <<
                                util::str(m_dev_count) << " requests";

    dsa->dS485()->removeListener(this);

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
    }
    else
    {
        m_task_status = Finished;
        m_status_info = _("Finished");
    }
    m_status_mutex.unlock();
}


TestOneDevicePresenceTask::TestOneDevicePresenceTask(dsuid_t dsuid,
        dsuid_t dsm_dsuid, int timeout) :
        dSLongTask(),
        m_dsuid(dsuid),
        m_dsm_dsuid(dsm_dsuid),
        m_result_err(0),
        m_timeout(timeout),
        m_waiting(0)
{
    m_name = _("Device presence");
    m_description = _P(_("Update device presence of %s"),
                                util::str(m_dsuid).c_str());
    m_parameters = "dsuid=" + util::str(dsuid) + "&dsm_dsuid=" + util::str(dsm_dsuid);
    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
    m_present = false;
}

TestOneDevicePresenceTask::~TestOneDevicePresenceTask()
{
    try
    {
        dSA::dS_Assistant::getInstance()->dS485()->removeListener(this);
    }
    catch(Exception& e)
    {
    }
}

task_type TestOneDevicePresenceTask::getTaskType() const
{
    return TestOneDevicePresence;
}

void TestOneDevicePresenceTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {
        default:
            break;

        case DsmApiDevicePing:
        {
            boost::shared_ptr<DevicePingData> data =
                    boost::dynamic_pointer_cast<DevicePingData> (busdata);
            m_present = data->getPresent();
            m_result_err = 0;
        }
        break;
    }
}

void TestOneDevicePresenceTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    if (kStateInit == m_state)
    {
        m_progress = 0;
        m_waiting = 0;
        try {
            m_device = Storage::getInstance()->getDevice(m_dsuid, m_dsm_dsuid);
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << "[Task OneDevice Presence] "
                    "Could not load device: " << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                FILE_LOG(logERROR) << "Unrecoverable SQLite error, terminating dSA";
                g_shutdown_flag = SHUTDOWN_SQLITE;
            }
        }
        catch (Exception &ex)
        {
            FILE_LOG(logINFO) << "[Task OneDevice Presence] "
                    "Device " << util::str(m_dsuid) <<
                    " on dSM " << util::str(m_dsm_dsuid) <<
                    " not found: " << ex.getMessage();
            m_status_mutex.lock();
            m_task_status = Finished;
            m_status_info = _P(_("Unknown device with dSUID %s"),
                                        util::str(m_dsuid).c_str());
            m_result_err = ex.getErrorCode();
            m_status_mutex.unlock();
            m_progress = 100;
            return;
        }

        dsa->dS485()->addListener(this);
        this->select(DsmApiDevicePing, true);

        m_state = kStateSetup;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid && (m_state != kStateFinished))
    {

        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                FILE_LOG(logINFO) << "[Task OneDevice Presence] "
                        "Device " << util::str(m_dsuid) <<
                        " on dSM " << util::str(m_dsm_dsuid);

                try
                {
                    m_result_err = ERROR_TIMEOUT;
                    dsa->dS485()->reqDevicePresenceTest(m_dsm_dsuid, m_device->getId());
                    m_state = kStateRequest;
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (shared_from_this()),
                            2);
                    return;
                }
                catch (Exception &ex)
                {
                    m_status_mutex.lock();
                    m_status_info = _P(
                            _("Failed to update presence for device %s: %s"),
                            util::str(m_dsuid).c_str(), ex.getMessage().c_str());
                    m_result_err = ex.getErrorCode();
                    m_status_mutex.unlock();

                    FILE_LOG(logINFO) << "[Task OneDevice Presence] "
                            "Device " << util::str(m_dsuid) <<
                            " on dSM " << util::str(m_dsm_dsuid) <<
                            ", error " << util::str(m_result_err) << " = "
                            << ex.getMessage();
                    m_state = kStateResponse;
                }
            }
            break;

            case kStateRequest:
            {
                if ((m_result_err == ERROR_TIMEOUT) && (m_waiting < m_timeout))
                {
                    m_waiting += 2;
                    dsa->scheduleLongTask(
                            boost::shared_ptr<dSLongTask> (shared_from_this()),
                            2);
                    m_status_info = _P(
                                    _("Waiting for response from device %s"),
                                    util::str(m_dsuid).c_str());
                    return;
                }
                m_state = kStateResponse;
            }
            break;

            case kStateResponse:
            {
                m_state = kStateFinished;
                m_status_mutex.lock();
                if (m_result_err == ERROR_TIMEOUT)
                {
                    m_status_info = _P(_("No response from device %s"),
                                                util::str(m_dsuid).c_str());
                }
                else if (m_result_err == ERROR_OK)
                {
                    m_status_info = _P(
                                        _("Received response from device %s"),
                                        util::str(m_dsuid).c_str());
                }
                else
                {
                    m_status_info = _P(
                                        _("Error when querying device %s: %s"),
                                        util::str(m_dsuid).c_str(),
                                        util::str(m_result_err).c_str());
                }
                m_progress = 100;
                m_status_mutex.unlock();

                FILE_LOG(logDEBUG1) << "[Task OneDevice Presence] "
                        "Device " << util::str(m_dsuid) <<
                        " on dSM " << util::str(m_dsm_dsuid) <<
                        " present: " << (m_present ? "yes" : "no");
            }
            break;

            case kStateFinished:
            {
                FILE_LOG(logWARNING) << "[Task OneDevice Presence] Force task end";
            }
            break;

            default:
            {
                m_state = kStateFinished;
            }
            break;

        } // switch
    }

    FILE_LOG(logINFO) << "[Task OneDevice Presence] Ready";

    dsa->dS485()->removeListener(this);

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
    }
    else
    {
        m_task_status = Finished;
    }
    m_status_mutex.unlock();
}

TestDevicePresenceTask::TestDevicePresenceTask(dsuid_t dsm_dsuid,
                                               int interval_time) :
        dSLongTask(),
        m_dsm_dsuid(dsm_dsuid),
        m_interval_time(interval_time),
        m_devs(new std::list<boost::shared_ptr<DeviceP> >),
        m_dsms(new std::list<boost::shared_ptr<MeterP> >),
        m_dsm_dev_count(0),
        m_total_dev_count(0),
        m_total_tx(0),
        m_total_rx(0),
        m_tx_count(0),
        m_tx_count_equal(0),
        m_tx_dev_count(0),
        m_rx_count(0)
{
    m_name = _("Device presence test");
    m_description = _("Test presence of registered devices");
    m_parameters = "dsm=" + util::str(dsm_dsuid) + '&' +
            "interval=" + util::str(interval_time);

    if (!dsuid_is_null(&dsm_dsuid))
    {
        m_description = _P(
                _("Test presence of registered devices on dSM %s"),
                util::str(dsm_dsuid).c_str());
    }
    else
    {
        m_description = _("Test presence of registered devices on all dSMs");
    }
    m_valid = true;
    m_pausable = true;
    m_state = 0;

    pthread_mutex_init(&m_ping_lock, NULL);
    pthread_cond_init(&m_ping_status_change, NULL);
}

TestDevicePresenceTask::~TestDevicePresenceTask()
{
    try
    {
        dSA::dS_Assistant::getInstance()->dS485()->removeListener(this);
    }
    catch(Exception& e)
    {
    }
    pthread_mutex_destroy(&m_ping_lock);
    pthread_cond_destroy(&m_ping_status_change);
}

task_type TestDevicePresenceTask::getTaskType() const
{
    return TestDevicePresence;
}

TestDevicePresenceTask::DeviceP::DeviceP(Device* parent) :
        Device(*parent),
        m_ping_state(pd_state_unknown)
{
    m_count_ping_sent = m_count_ping_received = 0;
}

TestDevicePresenceTask::MeterP::MeterP(dSM* parent, TestDevicePresenceTask* const task) :
        dSM(*parent), m_task(task), m_count_ping_timeout(0), m_a_selection(0),
        m_b_selection(0), m_d_selection(0)
{
    resetPingState();
}

void TestDevicePresenceTask::MeterP::resetPingState()
{
    m_count_ping_dut = m_count_ping_sent = m_count_ping_received = m_deviceid_dut = 0;
    m_state = pm_state_unknown;
}

void TestDevicePresenceTask::MeterP::getPingStatus(uint16_t* tx_dev_count,
                                                   uint16_t* tx_count,
                                                   uint16_t* rx_count)
{
    *tx_dev_count = m_count_ping_dut;
    *tx_count = m_count_ping_sent;
    *rx_count = m_count_ping_received;
}

void TestDevicePresenceTask::MeterP::signalPingSent(uint16_t device_id)
{
    if (device_id < 64)
    {
        return;
    }
    if (device_id != m_deviceid_dut)
    {
        m_deviceid_dut = device_id;
        m_count_ping_dut ++;
        m_count_ping_timeout = 0;

        boost::shared_ptr<DeviceP> dsd = m_task->getDevice(getDSUID(), device_id);
        if (dsd)
        {
            Database::getInstance()->updateDevicePresenceStatus(m_task->m_task_id,
                    util::str(dsd->getDSMdSUID()), device_id,
                    dsd->getPingState());
        }
    }
    m_state = pm_state_ping_sent;
    m_count_ping_sent ++;
    pthread_cond_signal(&m_task->m_ping_status_change);
}

void TestDevicePresenceTask::MeterP::signalPingReply(uint16_t device_id)
{
    if (device_id == m_deviceid_dut)
    {
        m_count_ping_received ++;
        m_state = pm_state_reply_received;

        boost::shared_ptr<DeviceP> dsd = m_task->getDevice(getDSUID(), device_id);
        if (dsd)
        {
            Database::getInstance()->updateDevicePresenceStatus(m_task->m_task_id,
                    util::str(dsd->getDSMdSUID()), device_id,
                    dsd->getPingState());
        }

        if (m_count_ping_dut >= getDeviceCount())
        {
            m_state = pm_state_ready;
        }
    }
    pthread_cond_signal(&m_task->m_ping_status_change);
}

void TestDevicePresenceTask::MeterP::signalPingTimeout()
{
    m_state = pm_state_waiting;
    if (++m_count_ping_timeout >= 3)
    {
        if (m_count_ping_dut >= getDeviceCount())
        {
            m_state = pm_state_ready;
        }
    }
}

boost::shared_ptr<TestDevicePresenceTask::MeterP>
                                    TestDevicePresenceTask::getDSM(dsuid_t dsmid)
{
    dsuid_t id = m_dsm_under_test->getDSUID();
    if (dsuid_equal(&id, &dsmid))
    {
        return m_dsm_under_test;
    }
    return boost::shared_ptr<TestDevicePresenceTask::MeterP>();
}

boost::shared_ptr<TestDevicePresenceTask::DeviceP>
            TestDevicePresenceTask::getDevice(dsuid_t dsmid, uint16_t devindex)
{
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::DeviceP> >::iterator itm = m_devs->begin();
            (itm != m_devs->end());
            itm++)
    {
        dsuid_t idm = (*itm)->getDSMdSUID();
        if (dsuid_equal(&idm, &dsmid) && ((*itm)->getId() == devindex))
        {
            return *itm;
        }
    }
    return boost::shared_ptr<TestDevicePresenceTask::DeviceP>();
}

boost::shared_ptr<TestDevicePresenceTask::DeviceP>
                TestDevicePresenceTask::getDevice(dsuid_t dsmid, uint32_t dev_serial)
{
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::DeviceP> >::iterator itm = m_devs->begin();
            (itm != m_devs->end());
            itm++)
    {
        dsuid_t idm = (itm->get())->getDSMdSUID();
        uint32_t idd = (itm->get())->getSerialNumber();
        if (dsuid_equal(&idm, &dsmid) && (idd == dev_serial))
        {
            return *itm;
        }
    }
    return boost::shared_ptr<TestDevicePresenceTask::DeviceP>();
}

void TestDevicePresenceTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {

        default:
            break;

        case DsmApiCommUpstreamShort:
            break;

        case DsmApiCommUpstreamLong:
        {
            boost::shared_ptr<CommUpstreamLongData> data =
                boost::dynamic_pointer_cast<CommUpstreamLongData> (busdata);
            if (data->getCrosstalkFlags() > 0)
            {
                return;
            }
            switch (data->getMainQualifier())
            {
                case 15:
                {
                    boost::shared_ptr<MeterP> dsm =
                            getDSM(data->getDSMSourceDSUID());
                    boost::shared_ptr<DeviceP> dsd =
                            getDevice(data->getDSMSourceDSUID(), data->getData());

                    std::string sourceId((dsd ? util::str(dsd->getId()) : "0000") +
                            "/" + util::hex(data->getData()));

                    FILE_LOG(logDEBUG2) << "Ping reply from " << sourceId;

                    if (dsd && dsd->getPingState() == pd_state_sent)
                    {
                        dsd->setPingState(pd_state_replied);
                        if (dsm)
                        {
                            dsm->signalPingReply(dsd->getId());
                        }
                    }
                }
                break;
            }
        }
        break;

        case DsmApiCommDownstreamShort:
        break;

        case DsmApiCommDownstreamLong0:
        {
            boost::shared_ptr<CommDownstreamLong0Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong0Data> (busdata));
            switch (data->getCommand())
            {
                case 0x02:
                {
                    boost::shared_ptr<MeterP> dsm =
                            getDSM(data->getDSMSourceDSUID());
                    if (dsm == NULL)
                    {
                        return;
                    }

                    boost::shared_ptr<DeviceP> dsd =
                            getDevice(data->getDSMSourceDSUID(), (uint16_t) (dsm->getSelectionB() >> 6));

                    std::string targetId(util::str(dsm->getSelectionB() >> 6) + "/" +
                            (dsd ? util::str(dsd->getDSUID()) : "00000000"));

                    FILE_LOG(logDEBUG2) << "Ping sent to " << targetId;
                    dsm->signalPingSent(dsm->getSelectionB() >> 6);
                    if (dsd)
                    {
                        if (dsd->getPingState() == pd_state_scheduled)
                        {
                            dsd->setPingState(pd_state_sent);
                        }
                    }
                } /* case 0x02 */
                break;
            } /* switch getCommand() */
        }
        break;

        case DsmApiCommDownstreamLong8:
        break;

        case DsmApiCommDownstreamLong16:
        {
            boost::shared_ptr<CommDownstreamLong16Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong16Data>(busdata));
            boost::shared_ptr<MeterP> dsm = getDSM(data->getDSMSourceDSUID());
            if (dsm)
            {
                static uint16_t dsuidh = 0;
                uint16_t payload = data->getData();
                switch (data->getCommand())
                {
                    case 0:
                        dsm->setSelectionA(payload);
                        break;
                    case 1:
                        dsm->setSelectionB(payload);
                        dsm->setSelectionD(0);
                        break;
                    case 2:
                        dsuidh = payload;
                        break;
                    case 3:
                        dsm->setSelectionD(
                                (uint32_t) dsuidh << 16 | data->getData());
                        dsuidh = 0;
                        break;
                }
            }
        }
        break;

        case DsmApiDeviceActive:
        {
            boost::shared_ptr<DeviceActiveData> data(
                    boost::dynamic_pointer_cast<DeviceActiveData> (busdata));
            boost::shared_ptr<DeviceP> dsd = getDevice(data->getDSMSourceDSUID(),
                                                       data->getDeviceID());
            if (dsd)
            {
                dsd->setPingState(pd_state_discovered);
            }
        }
        break;

        case DsmApiDeviceInactive:
        {
            boost::shared_ptr<DeviceInactiveData> data(
                    boost::dynamic_pointer_cast<DeviceInactiveData> (busdata));
            boost::shared_ptr<DeviceP> dsd = getDevice(data->getDSMSourceDSUID(),
                                                       data->getDeviceID());
            if (dsd)
            {
                dsd->setPingState(pd_state_inactive);
            }
        }
        break;
    }
}

void TestDevicePresenceTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_progress = 0;
    m_status_mutex.unlock();

    // initial setup
    if (m_state == 0)
    {
        boost::shared_ptr<std::list<boost::shared_ptr<Device> > > devs;
        boost::shared_ptr<std::list<dsuid_t> > dsms;
        boost::shared_ptr<std::list<dsuid_t> > dsm_list(new std::list<dsuid_t>);

        if (dsuid_is_null(&m_dsm_dsuid))
        {
            devs = Storage::getInstance()->getDevices(false);
            dsms = Storage::getInstance()->getDSMList();
        }
        else
        {
            devs = Storage::getInstance()->getDevices(m_dsm_dsuid, false);
            dsms = dsm_list;
            dsms->push_back(m_dsm_dsuid);
        }

        // copy device list and preset device ping state
        for (std::list<boost::shared_ptr<Device> >::iterator it =
                devs->begin();
                it != devs->end();
                it++)
        {
            boost::shared_ptr<DeviceP> dsd(new DeviceP((*it).get()));
            dsd->setPingState(pd_state_scheduled);
            m_devs->push_back(dsd);

            Database::getInstance()->saveDevicePresence(m_task_id,
                            util::str(dsd->getDSUID()),
                            util::str(dsd->getDSMdSUID()),
                            dsd->getId(),
                            dsd->getPingState());
        }

        // copy meter list and reset ping state
        m_total_dev_count = 0;
        for (std::list<dsuid_t>::iterator itm = dsms->begin();
                m_valid && (itm != dsms->end());
                itm++)
        {
            try
            {
                boost::shared_ptr<MeterP> dsm(
                        new MeterP(Storage::getInstance()->getDSM(*itm).get(), this));
                dsm->resetPingState();
                m_total_dev_count += dsm->getDeviceCount();
                m_dsms->push_back(dsm);
            }
            catch (SQLite3Exception &sx)
            {
                FILE_LOG(logERROR) << "Could not load dSM " <<
                    util::str(*itm) << ": " << sx.getMessage();
                if (SQLite3::isFatal(sx.getErrorCode()))
                {
                    throw sx;
                }
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << "[Task Device Presence] dSM with ID " <<
                        util::str(*itm) << " not found in database";
            }
        }

        if (0 == m_total_dev_count)
        {
            m_status_mutex.lock();
            m_progress = 100;
            m_task_status = Finished;
            m_status_mutex.unlock();
            return;
        }

        // add task to receive callbacks
        dsa->dS485()->addListener(this);

        // select the callbacks you want
        this->select(DsmApiCommUpstreamShort, true);
        this->select(DsmApiCommUpstreamLong, true);
        this->select(DsmApiCommDownstreamShort, true);
        this->select(DsmApiCommDownstreamLong0, true);
        this->select(DsmApiCommDownstreamLong8, true);
        this->select(DsmApiCommDownstreamLong16, true);
        this->select(DsmApiDeviceActive, true);
        this->select(DsmApiDeviceInactive, true);

        m_total_rx = m_total_tx = 0;
        m_tx_dev_count = m_tx_count = m_rx_count = 0;

        m_state ++;

        FILE_LOG(logINFO) << "[Task Device Presence] Starting presence check on " <<
                util::str(m_dsms->size()) << " meter(s) for a total of " <<
                util::str(m_total_dev_count) << " devices (db " <<
                util::str(m_devs->size()) << ")";
    }

    // resuming from paused state?
    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid && (m_state != 6))
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case 1:
            {
                if (m_dsms->empty())
                {
                    m_state = 6;
                }
                else
                {
                    m_dsm_under_test = m_dsms->front();
                    m_dsms->pop_front();
                    m_dsm_dev_count = m_dsm_under_test->getDeviceCount();
                    if (m_dsm_dev_count > 0)
                    {
                        m_status_mutex.lock();
                        m_status_info = "Check " + util::str(m_dsm_dev_count) + " devices on dsm " +
                                util::str(m_dsm_under_test->getDSUID());
                        m_status_mutex.unlock();
                        m_state ++;
                    }
                    m_rx_count = m_tx_count = 0;
                    m_tx_dev_count = 0;
                    m_tx_count_equal = 0;
                }
            }
            break;

            case 2:
            {
                uint8_t Hardware;
                uint8_t DSP;
                uint8_t Temperature;
                uint8_t Config;
                uint8_t dS485;
                uint8_t Line;
                try
                {
                    dS_Assistant::getInstance()->dS485()->dSMState(m_dsm_under_test->getDSUID(),
                            &Hardware, &DSP, &Temperature, &Config, &dS485, &Line);
                    if (Line)
                    {
                        FILE_LOG(logINFO) << "[Task Device Presence] dSM " <<
                                util::str(m_dsm_under_test->getDSUID()) << " is currently jumbling "
                                "and busy, device presence check skipped";
                        m_total_tx += m_dsm_dev_count;
                        m_state = 1;
                    }
                    else
                    {
                        m_state ++;
                    }
                }
                catch (Exception &ex)
                {
                    m_state = 1;
                }
            }
            break;

            case 3:
            {
                FILE_LOG(logDEBUG1) << "[Task Device Presence] Start to check " <<
                        util::str(m_dsm_dev_count) << " devices on dsm " <<
                        util::str(m_dsm_under_test->getDSUID());

                // send ping-one-time-off command, after that the 24h mode is re-enabled
                try
                {
                    dsa->dS485()->pingOneTimeOff(m_dsm_under_test->getDSUID(),
                            (uint16_t) m_interval_time);
                    m_state ++;
                }
                catch(Exception &ex)
                {
                    FILE_LOG(logERROR)
                            << "[Task Device Presence] dSM-API request failed: "
                            << ex.getMessage();
                    m_state = 1;
                }
            }
            break;

            case 4:
            {
                const uint16_t reschedule_time = 2;
                uint16_t tx_count_last = m_tx_count;
                uint16_t tx_threshold = (m_interval_time + 2500) / 1000;

                m_dsm_under_test->getPingStatus(&m_tx_dev_count, &m_tx_count, &m_rx_count);

                if (tx_count_last == m_tx_count)
                {
                    ++m_tx_count_equal;
                }
                else
                {
                    m_tx_count_equal = 0;
                }

                if (m_dsm_under_test->getState() == pm_state_ready ||
                        m_rx_count >= m_dsm_dev_count ||
                        m_tx_count_equal > tx_threshold)
                {
                    m_state = 1;
                    m_total_tx += m_tx_dev_count;
                    m_total_rx += m_rx_count;
                }
                else
                {
                    m_status_mutex.lock();
                    m_status_info = "Check " + util::str(m_dsm_dev_count) +
                            " devices on dsm " + util::str(m_dsm_under_test->getDSUID()) +
                            ", " + util::str(m_tx_dev_count) + " of " + util::str(m_dsm_dev_count) + " completed";
                    m_progress = (unsigned char)((m_total_tx + m_tx_dev_count) * 100 / m_total_dev_count);
                    m_status_mutex.unlock();

                    FILE_LOG(logDEBUG1) << "[Task Device Presence] Ping Progress = "
                            << util::str(m_progress) <<
                            ", tx-equal = " << util::str(m_tx_count_equal) <<
                            ", tx-total = " << util::str(m_total_tx) <<
                            ", tx-dev-count = " << util::str(m_tx_dev_count) <<
                            ", tx-count = " << util::str(m_tx_count) <<
                            ", rx-count = " << util::str(m_rx_count);

                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), reschedule_time);
                    return;
                }
            }
            break;

        }
    }

    // don't forget to unsubscribe from the listener
    dsa->dS485()->removeListener(this);

    // loop over devices and count successful replies
    size_t total_replies = 0;
    size_t total_inactive = 0;
    size_t total_discovered = 0;
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::DeviceP> >::iterator it = m_devs->begin();
            it != m_devs->end();
            it++)
    {
        switch ((*it)->getPingState())
        {
            case pd_state_replied:
                total_replies ++;
                break;
            case pd_state_inactive:
                Database::getInstance()->updateDevicePresenceStatus(m_task_id,
                        util::str((*it)->getDSMdSUID()), (*it)->getId(),
                        (*it)->getPingState());
                total_inactive ++;
                break;
            case pd_state_discovered:
                Database::getInstance()->updateDevicePresenceStatus(m_task_id,
                        util::str((*it)->getDSMdSUID()), (*it)->getId(),
                        (*it)->getPingState());
                total_discovered ++;
                break;
            default:
                break;
        }
    }

    FILE_LOG(logINFO) << "[Task Device Presence] Finished, "
            "total " << util::str(m_total_dev_count) << " devices checked, "
            "got " << util::str(total_replies) << " replies, " <<
            util::str(total_inactive) << " changed to inactive " <<
            util::str(total_discovered) << " changed to active";

    m_status_mutex.lock();
    m_progress = 100;
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
    }
    else
    {
        m_task_status = Finished;
        m_status_info = _("Finished");
    }
    m_status_mutex.unlock();
}

TestDeviceStatisticTask::TestDeviceStatisticTask(dsuid_t dsm_dsuid,
        dsuid_t dev_dsuid) :
        dSLongTask(), m_dsm_dsuid(dsm_dsuid), m_dev_dsuid(dev_dsuid),
        m_offset(0), m_length(0), m_retry(0), m_result_err(0),
        m_result_value(0), m_dev_count(0)
{
    m_name = _("Statistics data of a device");
    m_description = _P(_("Query statistics data from %s"),
                       util::str(dev_dsuid).c_str());
    m_parameters = "dsm=" + util::str(dsm_dsuid) + "&dsuid=" +
                                                        util::str(dev_dsuid);
    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
}

TestDeviceStatisticTask::~TestDeviceStatisticTask()
{
    try
    {
        dSA::dS_Assistant::getInstance()->dS485()->removeListener(this);
    }
    catch(Exception& e)
    {
    }
}

task_type TestDeviceStatisticTask::getTaskType() const
{
    return TestDeviceStatistic;
}

void TestDeviceStatisticTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {
        default:
            break;

        case DsmApiDeviceStatistics:
        {
            boost::shared_ptr<DeviceStatisticData> data =
                boost::dynamic_pointer_cast<DeviceStatisticData> (busdata);

            dsuid_t idm = data->getDSMSourceDSUID();
            dsuid_t idd = m_device->getDSMdSUID();
            if (dsuid_equal(&idm, &idd) &&
                (data->getDeviceID() == m_device->getId()))
            {
                m_result_value = data->getValue();
                m_result_err = 0;
                m_pstats->setValue(data->getOffset() * 2, m_result_value, 0);
                m_retry = 0;
            }
        }
        break;
    }
}

void TestDeviceStatisticTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    if (m_state == kStateInit)
    {
        try
        {
            if (dsuid_is_null(&m_dsm_dsuid))
            {
                m_dev_list = Storage::getInstance()->getDevices(true);
            }
            else if (dsuid_is_null(&m_dev_dsuid))
            {
                m_dev_list = Storage::getInstance()->getDevices(m_dsm_dsuid, true);
            }
            else
            {
                m_dev_list = boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                    (new std::list<boost::shared_ptr<Device> >);
                m_dev_list->push_back(
                        Storage::getInstance()->getDevice(m_dev_dsuid, m_dsm_dsuid));
            }
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                throw sx;
            }
            m_dev_list.reset();
        }
        catch (Exception &ex)
        {
            m_dev_list.reset();
        }
        m_dev_count = 0;
        m_dev_current = m_dev_list->begin();
        m_progress = 0;

        dsa->dS485()->addListener(this);
        this->select(DsmApiDeviceStatistics, true);

        m_state = kStateSetup;
    }

    if (!m_dev_list || m_dev_list->size() == 0)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid && (m_state != kStateFinished))
    {
        if (m_paused)
        {
            m_task_status = Sleeping;
            return;
        }

        if (!dsa->getLinkState())
        {
            FILE_LOG(logINFO) << "[Task Device Statistic] "
                    "Suspended, dS485 bus down";
            m_task_status = Sleeping;
            dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 30);
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                m_device = *m_dev_current;
                m_dev_count ++;
                m_offset = 0;
                m_length = 0;
                m_retry = 0;
                m_pstats.reset();
                m_status_info = _P(
                                    _("Requesting statistic from device %s"),
                                    util::str(m_device->getDSUID()).c_str());
                FILE_LOG(logINFO) << "[Task Device Statistic] "
                        "Start to query statistic values from device " <<
                        util::str(m_device->getDSUID());
                m_state = kStateVersion;
            }
            break;

            case kStateVersion:
            {
                m_retry = 0;
                while (m_retry < 3)
                {
                    try
                    {
                        dsa->dS485()->getStatisticValue(m_device->getDSMdSUID(),
                                m_device->getId(), m_offset, &m_result_value);
                        m_pstats = boost::shared_ptr<StatisticDevice>
                            (new StatisticDevice(m_result_value & 0xff));
                        m_pstats->pushValue(m_result_value, 0);
                        m_length = m_result_value >> 9;
                        m_state = kStateRequest;
                        m_offset ++;
                        m_retry = 0;
                        break;
                    }
                    catch (Exception &ex)
                    {
                        m_result_err = ex.getErrorCode();
                        m_retry ++;
                    }
                }
                if (m_retry)
                {
                    FILE_LOG(logINFO) << "[Task Device Statistic] "
                            "Failed to get statistic version from device " <<
                            util::str(m_device->getDSUID()) <<
                            " [Error " << util::str(m_result_err) << "]";
                    Database::getInstance()->saveStatisticDevice(m_task_id,
                            m_device->getDSUID(), m_device->getDSMdSUID(),
                            m_device->getId(), STAT_VERSION, 0, m_result_err);
                    m_state = kStateNext;
                }
            }
            break;

            case kStateRequest:
            {
                m_result_err = ERROR_TIMEOUT;
                m_progress = (unsigned char)
                        (((m_dev_count-1) * 54 + m_offset) * 100 / (m_dev_list->size() * 54));
                try
                {
                    dsa->dS485()->reqStatisticValue(m_device->getDSMdSUID(),
                            m_device->getId(), m_offset);
                    m_state = kStateResponse;
                }
                catch (Exception &ex)
                {
                    m_result_err = ex.getErrorCode();
                    m_retry ++;
                }

                if (m_retry >= 3)
                {
                    FILE_LOG(logINFO) << "[Task Device Statistic] "
                            "Failed to get offset " <<
                            util::str(m_offset) << " from device " <<
                            util::str(m_device->getDSUID()) <<
                            " [Error " << util::str(m_result_err) << "]";
                    m_pstats->setValue(m_offset * 2, 0, m_result_err);
                    m_offset ++;
                    m_retry = 0;
                }
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 5);
                return;
            }
            break;

            case kStateResponse:
            {
                m_state = kStateRequest;
                if (m_result_err)
                {
                    if ((m_retry++ >= 3) || (m_result_err != ERROR_TIMEOUT))
                    {
                        FILE_LOG(logINFO) << "[Task Device Statistic] "
                                "Failed to get offset " <<
                                util::str(m_offset) << " from device " <<
                                util::str(m_device->getDSUID()) <<
                                " [Error " << util::str(m_result_err) << "]";
                        m_pstats->setValue(m_offset * 2, 0, m_result_err);
                        m_offset ++;
                        m_retry = 0;
                    }
                    continue;
                }
                if (++m_offset >= m_length)
                {
                    m_state = kStateReady;
                }
                if (m_dev_list->size() > 0)
                {
                    m_progress = (unsigned char)
                            (((m_dev_count-1) * 54 + m_offset) * 100 / (m_dev_list->size() * 54));
                }
            }
            break;

            case kStateReady:
            {
                m_pstats->parseValues();

                FILE_LOG(logINFO) << "[Task Device Statistic] "
                        "Query complete got " <<
                        util::str(m_offset) << " values from device " <<
                        util::str(m_device->getDSUID());

                Database::getInstance()->saveStatisticDevice(m_task_id,
                        m_device->getDSUID(), m_device->getDSMdSUID(),
                        m_device->getId(), m_pstats->m_value_map);

                DSUID_PREFIX_LOG(logINFO, m_device->getDSUID(), "stat_")
                    << "Statistic Query Results:";

                for (std::list<boost::shared_ptr<StatisticValue> >::iterator it =
                        m_pstats->m_value_map->begin();
                        it != m_pstats->m_value_map->end();
                        it++)
                {
                    std::pair<std::string,std::string> desc =
                            Database::getInstance()->getStatisticDescription((*it)->m_field);
                    if ((*it)->m_errorcode)
                    {
                        DSUID_PREFIX_LOG(logINFO, m_device->getDSUID(), "stat_")
                            << desc.first << " = ERROR " << (*it)->m_errorcode;
                    }
                    else
                    {
                        DSUID_PREFIX_LOG(logINFO, m_device->getDSUID(), "stat_")
                            << desc.first << " = " << m_pstats->formatValue(*it);
                    }
                }

                m_state = kStateNext;
            }
            break;

            case kStateNext:
            {
                m_state = kStateSetup;
                m_dev_current ++;
                if (m_dev_current == m_dev_list->end())
                {
                    m_state = kStateFinished;
                }
            }
            break;

            default:
                m_state = kStateFinished;
                break;
        }
    }

    dsa->dS485()->removeListener(this);

    m_status_mutex.lock();
    m_progress = 100;
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
    }
    else
    {
        m_task_status = Finished;
        m_status_info = _("Finished");
    }
    m_status_mutex.unlock();
}

TestDeviceStatisticResetTask::TestDeviceStatisticResetTask(dsuid_t dsm_dsuid,
        dsuid_t dev_dsuid) :
        dSLongTask(), m_dsm_dsuid(dsm_dsuid), m_dev_dsuid(dev_dsuid),
        m_retry(0), m_errorcode(0), m_dev_count(0)
{
    m_name = _("Reset statistics data");
    m_description = _P(_("Reset statistics of device %s"),
                                util::str(dev_dsuid).c_str());
    m_parameters = "dsm=" + util::str(dsm_dsuid) + "&dsuid=" +
                                                           util::str(dev_dsuid);
    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
}

task_type TestDeviceStatisticResetTask::getTaskType() const
{
    return TestDeviceStatisticReset;
}

void TestDeviceStatisticResetTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    if (m_state == kStateInit)
    {
        try
        {
            if (dsuid_is_null(&m_dsm_dsuid))
            {
                m_dev_list = Storage::getInstance()->getDevices(true);
            }
            else if (dsuid_is_null(&m_dev_dsuid))
            {
                m_dev_list = Storage::getInstance()->getDevices(m_dsm_dsuid, true);
            }
            else
            {
                m_dev_list = boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                    (new std::list<boost::shared_ptr<Device> >);
                m_dev_list->push_back(
                        Storage::getInstance()->getDevice(m_dev_dsuid, m_dsm_dsuid));
            }
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                throw sx;
            }
            m_dev_list.reset();
        }
        catch (Exception &ex)
        {
            m_dev_list.reset();
        }
        m_dev_count = 0;
        m_dev_current = m_dev_list->begin();
        m_progress = 0;
        m_state = kStateSetup;
    }

    if (!m_dev_list || m_dev_list->size() == 0)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid && m_state != kStateFinished)
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                m_device = *m_dev_current;
                m_dev_count ++;
                m_retry = 0;
                m_state = kStateRequest;
                if (m_dev_list->size() > 0)
                {
                    m_progress = (unsigned char) ((m_dev_count-1) * 100 / m_dev_list->size());
                }
                m_status_info = _P(
                                    _("Reset statistic values of device %s"),
                                    util::str(m_device->getDSUID()).c_str());
                FILE_LOG(logINFO) << "[Task Device Statistic Reset] "
                        "Reset statistic values of device " <<
                        util::str(m_device->getDSUID());
            }
            break;

            case kStateRequest:
            {
                m_retry = 0;
                m_errorcode = 0;
                while (m_retry < 3)
                {
                    try
                    {
                        dsa->dS485()->reqStatisticValue(m_device->getDSMdSUID(),
                                m_device->getId(), 254);
                        m_state = kStateResponse;
                        m_retry = 0;
                        break;
                    }
                    catch (Exception &ex)
                    {
                        m_errorcode = ex.getErrorCode();
                        m_retry ++;
                    }
                }
                if (m_retry)
                {
                    FILE_LOG(logINFO) << "[Task Device Statistic Reset] "
                            "Failed to reset statistic version from device " <<
                            util::str(m_device->getDSUID()) <<
                            " [Error " << util::str(m_errorcode) << "]";
                    m_state = kStateNext;
                }
            }
            break;

            case kStateResponse:
            {
                m_state = kStateNext;
                if (m_errorcode != 0)
                {
                    m_state = kStateRequest;
                }
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 3);
                return;
            }
            break;

            case kStateNext:
            {
                m_state = kStateSetup;
                m_dev_current ++;
                if (m_dev_current == m_dev_list->end())
                {
                    m_state = kStateFinished;
                }
            }
            break;

            default:
                m_state = kStateFinished;
                break;
        }
    }

    m_status_mutex.lock();
    m_progress = 100;
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
    }
    else
    {
        m_task_status = Finished;
        m_status_info = _("Finished");
    }
    m_status_mutex.unlock();
}

TestMeterStatisticTask::TestMeterStatisticTask(dsuid_t dsm_dsuid) :
        dSLongTask(), m_dsm_dsuid(dsm_dsuid)
{
    m_name = _("Statistics data of a meter");
    m_description = _P(_("Query statistics data from meter %s"),
                                util::str(m_dsm_dsuid).c_str());
    m_parameters = "dsm=" + util::str(m_dsm_dsuid);
    m_valid = true;
}

task_type TestMeterStatisticTask::getTaskType() const
{
    return TestMeterStatistic;
}

void TestMeterStatisticTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    boost::shared_ptr<std::list<dsuid_t> > dsms;
    boost::shared_ptr<std::list<dsuid_t> > dsm_list(new std::list<dsuid_t>);

    if (dsuid_is_null(&m_dsm_dsuid))
    {
        dsms = Storage::getInstance()->getDSMList();
    }
    else
    {
        dsms = dsm_list;
        dsms->push_back(m_dsm_dsuid);
    }

    size_t total = dsms->size();
    size_t current = 0;
    if (total == 0)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    // loop over all dSMeter
    for (std::list<dsuid_t>::iterator itm = dsms->begin();
            m_valid && (itm != dsms->end());
            itm++)
    {
        current ++;
        m_status_mutex.lock();
        m_progress = (unsigned char)(current * 100 / total);
        m_status_info = _P(_("Query from device %zu of %zu"),
                                    current, total);
        m_status_mutex.unlock();

        FILE_LOG(logINFO) << "[Task Meter Statistic] Started query from dsm " <<
                util::str(*itm);

        uint8_t length;
        uint8_t index;
        uint8_t data[256];
        try
        {
            length = index = 0;
            dsa->dS485()->getStatisticUpstream(*itm, 0, &length, data + index);
            index += length;
            dsa->dS485()->getStatisticUpstream(*itm, 1, &length, data + index);
            index += length;
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR) << "[Task Meter Statistic] Query from dsm " <<
                    util::str(*itm) << " failed: " << ex.getMessage();
            continue;
        }
        uint16_t sversion = *(uint16_t *) data;
        boost::shared_ptr<StatisticMeter> pstats(new StatisticMeter(sversion));

        for (int i = 0; i < index; i++)
        {
            pstats->pushValue(data[i], 0);
        }
        pstats->parseValues();

        Database::getInstance()->saveStatisticMeter(m_task_id, *itm, pstats->m_value_map);

        boost::shared_ptr<std::map<statistic_field_t, std::string > > sresults =
            pstats->formatValues();
        for (std::map<statistic_field_t, std::string>::iterator it = sresults->begin();
             it != sresults->end();
             it++)
        {
            std::pair<std::string,std::string> desc =
                    Database::getInstance()->getStatisticDescription(
                                                                (*it).first);
            DSUID_PREFIX_LOG(logINFO, *itm, "stat_") << desc.first
                                                    << " = " << (*it).second;
        }
    }

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
    }
    else
    {
        m_task_status = Finished;
        m_status_info = _("Finished");
    }
    m_status_mutex.unlock();
}

TestMeterStatisticResetTask::TestMeterStatisticResetTask(dsuid_t dsm_dsuid) :
        dSLongTask(), m_dsm_dsuid(dsm_dsuid)
{
    m_name = _("Reset meter statistics");
    m_description = _P(_("Reset statistics data for dSM %s"),
                                util::str(m_dsm_dsuid).c_str());
    m_parameters = "dsm=" + util::str(m_dsm_dsuid);
    m_valid = true;
}

task_type TestMeterStatisticResetTask::getTaskType() const
{
    return TestMeterStatisticReset;
}

void TestMeterStatisticResetTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    boost::shared_ptr<std::list<dsuid_t> > dsms;
    boost::shared_ptr<std::list<dsuid_t> > dsm_list(new std::list<dsuid_t>);

    if (dsuid_is_null(&m_dsm_dsuid))
    {
        dsms = Storage::getInstance()->getDSMList();
    }
    else
    {
        dsms = dsm_list;
        dsms->push_back(m_dsm_dsuid);
    }

    size_t total = dsms->size();
    size_t current = 0;
    if (total == 0)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    // loop over all dSMeter
    for (std::list<dsuid_t>::iterator itm = dsms->begin();
            m_valid && (itm != dsms->end());
            itm++)
    {
        current ++;
        m_status_mutex.lock();
        m_progress = (unsigned char)(current * 100 / total);
        m_status_info = _P(_("Statistics reset, device: %zu of %zu"),
                                   current, total);
        m_status_mutex.unlock();

        try
        {
            dsa->dS485()->resetStatisticUpstream(*itm);
        }
        catch (Exception &ex)
        {
            FILE_LOG(logERROR) << "[Task Meter Statistic Reset] "
                    "Command failed for meter " <<
                    util::str(*itm) << ": " << ex.getMessage();
            continue;
        }
    }

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
    }
    else
    {
        m_task_status = Finished;
        m_status_info = _("Finished");
    }
    m_status_mutex.unlock();
}

TestSystemMonitorTask::TestSystemMonitorTask(int profile, int duration) :
        dSLongTask(), m_profile(profile), m_duration(duration)
{
    m_name = _("System Performance Monitor Test");
    m_description = _("Periodic query of communication and statistics data");
    m_parameters = "profile=" + util::str(m_profile) + ";duration=" + util::str(m_duration);
    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
}

task_type TestSystemMonitorTask::getTaskType() const
{
    return TestSystemMonitor;
}

TestSystemMonitorTask::SystemTestMethod::SystemTestMethod(
        task_type task, dsuid_t dsm_dsuid, dsuid_t device_dsuid, time_t dt) :
        m_type(task), m_dsm_dsuid(dsm_dsuid), m_device_dsuid(device_dsuid), m_deltatime(dt)
{
    m_lasttime = time(NULL);
}

dSLongTask* TestSystemMonitorTask::SystemTestMethod::getTask()
{
    return m_task.get();
}

std::string TestSystemMonitorTask::SystemTestMethod::getName()
{
    std::string result;
    switch (m_type)
    {
        case TestTransmissionQuality:
            result = "Systemtest Transmission Quality";
            break;
        case TestDevicePresence:
            result = "Systemtest Device Presence";
            break;
        case TestMeterStatistic:
            result = "Systemtest Meter Statistic";
            break;
        case TestDeviceStatistic:
            result = "Systemtest Device Statistic";
            break;
        case TestDeviceStatisticReset:
            result = "Systemtest Device Statistic Reset";
            break;
        case TestMeterStatisticReset:
            result = "Systemtest Meter Statistic Reset";
            break;
        default:
            result = "Invalid Systemtest Task";
            break;
    }
    return result;
}

bool TestSystemMonitorTask::SystemTestMethod::isRunning()
{
    return (m_task && ((m_task->getTaskStatus() == Running) ||
      (m_task->getTaskStatus() == Queued) ||
      (m_task->getTaskStatus() == Sleeping)));
}

bool TestSystemMonitorTask::SystemTestMethod::isToBeScheduled()
{
    if (m_deltatime == 0 || m_deltatime == -1)
    {
        return false;
    }
    return (time(NULL) > m_lasttime + m_deltatime);
}

long long int TestSystemMonitorTask::SystemTestMethod::schedule(long long int parent_id)
{
    if (m_task && m_task.use_count())
    {
        m_task.reset();
    }
    switch (m_type)
    {
        case TestTransmissionQuality:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestTranmissionQualityTask(1, m_dsm_dsuid));
            break;
        case TestDevicePresence:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestDevicePresenceTask(m_dsm_dsuid, 2000));
            break;
        case TestMeterStatistic:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestMeterStatisticTask(m_dsm_dsuid));
            break;
        case TestMeterStatisticReset:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestMeterStatisticResetTask(m_dsm_dsuid));
            break;
        case TestDeviceStatistic:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestDeviceStatisticTask(m_dsm_dsuid, m_device_dsuid));
            break;
        case TestDeviceStatisticReset:
            m_task = boost::shared_ptr<dSLongTask>
                (new TestDeviceStatisticResetTask(m_dsm_dsuid, m_device_dsuid));
            break;
        default:
            return 0;
    }
    m_task->setParentTaskID(parent_id);
    m_lasttime = time(NULL);
    return dS_Assistant::getInstance()->addLongTask(m_task);
}

void TestSystemMonitorTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    // initial setup
    if (kStateInit == m_state)
    {
        m_startup_methods = boost::shared_ptr<std::list<boost::shared_ptr<SystemTestMethod> > >
            (new std::list<boost::shared_ptr<SystemTestMethod> >);
        m_test_methods = boost::shared_ptr<std::list<boost::shared_ptr<SystemTestMethod> > >
            (new std::list<boost::shared_ptr<SystemTestMethod> >);
        m_cleanup_methods = boost::shared_ptr<std::list<boost::shared_ptr<SystemTestMethod> > >
            (new std::list<boost::shared_ptr<SystemTestMethod> >);

        switch (m_profile)
        {
            default:
            case 1:
            {
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatisticReset, DSUID_NULL, DSUID_NULL, 0)));
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatisticReset, DSUID_NULL, DSUID_NULL, 0)));

                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestTransmissionQuality, DSUID_NULL, DSUID_NULL, 10 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDevicePresence, DSUID_NULL, DSUID_NULL, 20 * 60)));

                m_cleanup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, DSUID_NULL, DSUID_NULL, 0)));
                m_cleanup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, DSUID_NULL, DSUID_NULL, 0)));
            }
            break;

            case 2:
            {
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatisticReset, DSUID_NULL, DSUID_NULL, 0)));
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatisticReset, DSUID_NULL, DSUID_NULL, 0)));

                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestTransmissionQuality, DSUID_NULL, DSUID_NULL, 1 * 60 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDevicePresence, DSUID_NULL, DSUID_NULL, 6 * 60 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, DSUID_NULL, DSUID_NULL, 1 * 60 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, DSUID_NULL, DSUID_NULL, 24 * 60 * 60)));
            }
            break;

            case 3:
            {
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatisticReset, DSUID_NULL, DSUID_NULL, 0)));
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatisticReset, DSUID_NULL, DSUID_NULL, 0)));

                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestTransmissionQuality, DSUID_NULL, DSUID_NULL, 15 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDevicePresence, DSUID_NULL, DSUID_NULL, 12 * 60 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, DSUID_NULL, DSUID_NULL, 20 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, DSUID_NULL, DSUID_NULL, 6 * 60 * 60)));
            }
            break;

            case 4:
            {
                m_startup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatisticReset, DSUID_NULL, DSUID_NULL, 0)));

                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestTransmissionQuality, DSUID_NULL, DSUID_NULL, 15 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDevicePresence, DSUID_NULL, DSUID_NULL, 20 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, DSUID_NULL, DSUID_NULL, 15 * 60)));
                m_test_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, DSUID_NULL, DSUID_NULL, 45 * 60)));

                m_cleanup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestMeterStatistic, DSUID_NULL, DSUID_NULL, 0)));
                m_cleanup_methods->push_back(boost::shared_ptr<SystemTestMethod> (
                        new SystemTestMethod(TestDeviceStatistic, DSUID_NULL, DSUID_NULL, 0)));
            }
            break;

        }

        m_state = kStateStartup;

        FILE_LOG(logINFO) << "[Profile Task] starting profile " <<
                util::str(m_profile) << " for " <<
                util::str(m_duration) << " seconds";
    }

    // resuming from paused state?
    if (m_paused)
    {
        m_paused = false;
    }

    // times for next execution of this control handler
    time_t next_schedule = 60;
    time_t next_task_schedule;

    // worker loop
    while (m_valid && (m_state != kStateFinished))
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateStartup:
            {
                // schedule tasks to run once at the begin of the tests
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_startup_methods->begin();
                        it != m_startup_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);
                    FILE_LOG(logDEBUG3) << "[Profile Task] schedule startup task: " << tm->getName();
                    tm->schedule(getTaskID());
                }
                m_state = kStateStartupWait;
                m_task_status = Sleeping;
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 30);
                return;
            }
            break;

            case kStateStartupWait:
            {
                int active_tasks = 0;
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_startup_methods->begin();
                        it != m_startup_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);
                    if (tm->isRunning())
                    {
                        active_tasks ++;
                    }
                }
                if (active_tasks == 0)
                {
                    m_state = kStateReady;
                }
                else
                {
                    m_task_status = Sleeping;
                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 30);
                    return;
                }
            }
            break;

            case kStateReady:
            {
                next_schedule = 60;
                time_t now = time(NULL);
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_test_methods->begin();
                        it != m_test_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);

                    next_task_schedule = 60;
                    if (tm->isRunning())
                    {
                        FILE_LOG(logDEBUG3)
                            << "[Profile Task] still running: "
                            << tm->getName();
                        next_task_schedule = 10;
                    }
                    else if (now >= tm->m_lasttime + tm->m_deltatime)
                    {
                        FILE_LOG(logDEBUG3)
                            << "[Profile Task] schedule subtask: "
                            << tm->getName();
                        tm->schedule(getTaskID());
                    }
                    else
                    {
                        next_task_schedule = tm->m_lasttime + tm->m_deltatime - now;
                        FILE_LOG(logDEBUG3)
                            << "[Profile Task] subtask " << tm->getName()
                            << " to be scheduled in "
                            << util::str(next_task_schedule);
                    }
                    if (next_schedule > next_task_schedule)
                    {
                        next_schedule = next_task_schedule;
                    }
                }
                m_state = kStateWait;
            }
            break;

            case kStateWait:
            {
                m_state = kStateReady;
                time_t now = time(NULL);
                if (now - m_started >= m_duration)
                {
                    FILE_LOG(logDEBUG3)
                        << "[Profile Task] main test time over";;
                    m_state = kStateCleanup;
                }
                else if (next_schedule > 0)
                {
                    m_task_status = Sleeping;
                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), next_schedule);
                    FILE_LOG(logDEBUG3)
                        << "[Profile Task] sleep for "
                        << util::str(next_schedule);
                    return;
                }
            }
            break;

            case kStateCleanup:
            {
                int active_tasks = 0;
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_test_methods->begin();
                        it != m_test_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);
                    if (tm->isRunning())
                    {
                        active_tasks ++;
                    }
                }
                if (active_tasks == 0)
                {
                    // schedule tasks to run once at the end of the tests
                    for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                            it = m_cleanup_methods->begin();
                            it != m_cleanup_methods->end();
                            it++)
                    {
                        boost::shared_ptr<SystemTestMethod> tm(*it);
                        FILE_LOG(logDEBUG3) << "[Profile Task] schedule cleanup task: " << tm->getName();
                        tm->schedule(getTaskID());
                    }
                    m_state = kStateCleanupWait;
                }
                m_task_status = Sleeping;
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 10);
                return;
            }
            break;

            case kStateCleanupWait:
            {
                int active_tasks = 0;
                for (std::list<boost::shared_ptr<SystemTestMethod> >::iterator
                        it = m_cleanup_methods->begin();
                        it != m_cleanup_methods->end();
                        it++)
                {
                    boost::shared_ptr<SystemTestMethod> tm(*it);
                    if (tm->isRunning())
                    {
                        active_tasks ++;
                    }
                }
                if (active_tasks == 0)
                {
                    m_state = kStateFinished;
                }
                else
                {
                    m_task_status = Sleeping;
                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 10);
                    return;
                }
            }
            break;

            default:
                m_state = kStateFinished;
                break;
        }
    }

    FILE_LOG(logINFO) << "[Profile Task] ready";

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
    }
    else
    {
        m_task_status = Finished;
        m_status_info = _("Finished");
    }
    m_status_mutex.unlock();
}


BasicMonitorTask::BasicMonitorTask() : dSTask()
{
    m_name = _("Basic traffic monitor");
    m_description = _("Permanently monitor and log ds traffic");
    m_valid = true;
    m_state = kStateInit;
}

BasicMonitorTask::~BasicMonitorTask()
{
    try
    {
        dSA::dS_Assistant::getInstance()->dS485()->removeListener(this);
    }
    catch(Exception& e)
    {
    }
}

void BasicMonitorTask::log(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {
        default:
            break;

        case DsmApiCommUpstreamShort:
        {
            boost::shared_ptr<CommUpstreamShortData> data =
                boost::dynamic_pointer_cast<CommUpstreamShortData> (busdata);

            std::string decoded = dsutil::rkDecodeShort(
                    data->getDSMSourceDSUID(),
                    data->getDeviceID(),
                    data->isResend(),
                    data->isSensor(),
                    data->getKeyType(),
                    data->getClickType(),
                    data->getFlags(),
                    data->getCrosstalkFlags());

            DSUID_LOG_TS(logINFO, data->getDSMSourceDSUID(), data->getTimestamp()) <<
                "UP-S [Q:" << std::setprecision (2) << (data->getQuality() * 100.0/62 + .5) <<
                ", P:" << ((data->getFlags() & 1) > 0 ? 'N' : 'P') <<
                ", C:" << ((data->getCrosstalkFlags() & 1) > 0 ? '1' : '0') <<
                "] " << decoded;
        }
        break;

        case DsmApiCommUpstreamLong:
        {
            boost::shared_ptr<CommUpstreamLongData> data =
                boost::dynamic_pointer_cast<CommUpstreamLongData> (busdata);

            std::string decoded = dsutil::rkDecodeLong(
                    data->getDSMSourceDSUID(),
                    data->getDeviceID(),
                    data->getMainQualifier(),
                    data->getSubQualifier(),
                    data->getData(),
                    data->getFlags(),
                    data->getCrosstalkFlags());

            DSUID_LOG_TS(logINFO, data->getDSMSourceDSUID(), data->getTimestamp()) <<
                "UP-L [Q:" << std::setprecision (2) << (data->getQuality() * 100.0/62 + .5) <<
                ", P:" << ((data->getFlags() & 1) > 0 ? 'N' : 'P') <<
                ", C:" << ((data->getCrosstalkFlags() & 1) > 0 ? '1' : '0') <<
                "] " << decoded;
        }
        break;

        case DsmApiCommDownstreamShort:
        {
            boost::shared_ptr<CommDownstreamShortData> data =
                boost::dynamic_pointer_cast<CommDownstreamShortData> (busdata);

            DSUID_LOG_TS(logINFO, data->getDSMSourceDSUID(), data->getTimestamp()) <<
                        dsutil::hkDecodeShort(data->getCommand());
        }
        break;

        case DsmApiCommDownstreamLong0:
        {
            boost::shared_ptr<CommDownstreamLong0Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong0Data> (busdata));

            DSUID_LOG_TS(logINFO, data->getDSMSourceDSUID(), data->getTimestamp()) <<
                        dsutil::hkDecodeLong0(data->getCommand());
        }
        break;

        case DsmApiCommDownstreamLong8:
        {
            boost::shared_ptr<CommDownstreamLong8Data> data(
                    boost::dynamic_pointer_cast<CommDownstreamLong8Data> (busdata));

            DSUID_LOG_TS(logINFO, data->getDSMSourceDSUID(), data->getTimestamp()) <<
                        dsutil::hkDecodeLong8(data->getCommand(), data->getData());
        }
        break;

        case DsmApiCommDownstreamLong16:
        {
            boost::shared_ptr<CommDownstreamLong16Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong16Data>(busdata));

            DSUID_LOG_TS(logINFO, data->getDSMSourceDSUID(), data->getTimestamp()) <<
                        dsutil::hkDecodeLong16(data->getCommand(), data->getData());
        }
        break;

    }

}

void BasicMonitorTask::callback(boost::shared_ptr<CBData> busdata)
{
    boost::shared_ptr<CBData> data = busdata;

    m_mutex.lock();
    m_queue.push(data);
    m_mutex.unlock();
}

void BasicMonitorTask::process()
{
    boost::shared_ptr<CBData> busdata;
    boost::shared_ptr<std::list<boost::shared_ptr<CBData> > > cbdatalist(
            new(std::list<boost::shared_ptr<CBData> >));

    m_mutex.lock();
    while (!m_queue.empty())
    {
        busdata = m_queue.front();
        m_queue.pop();
        cbdatalist->push_back(busdata);
    }
    m_mutex.unlock();

    for (std::list<boost::shared_ptr<CBData> >::iterator it = cbdatalist->begin();
            it != cbdatalist->end();
            it++)
    {
        log(*it);
    }
}

void BasicMonitorTask::run(dS_Assistant *dsa)
{
    // initial setup
    if (kStateInit == m_state)
    {
        // add task to receive callbacks
        dsa->dS485()->addListener(this);

        // select the callbacks you want
        this->select(DsmApiCommUpstreamShort, true);
        this->select(DsmApiCommUpstreamLong, true);
        this->select(DsmApiCommDownstreamShort, true);
        this->select(DsmApiCommDownstreamLong0, true);
        this->select(DsmApiCommDownstreamLong8, true);
        this->select(DsmApiCommDownstreamLong16, true);

        m_state = kStateSetup;
    }

    while (m_valid)
    {
        switch (m_state)
        {
            case kStateSetup:
            {
                process();
                m_state = kStateReady;
            }
            break;

            case kStateReady:
            default:
            {
                m_state = kStateSetup;
                dsa->scheduleTask(boost::shared_ptr<dSTask> (shared_from_this()), 5);
                return;
            }
            break;
        }
    }

    // final cleanup
    dsa->dS485()->removeListener(this);
}

DsmMonitorTask::DsmMonitorTask() : dSTask()
{
    m_name = _("dSMeter status monitor");
    m_description = _("Permanently monitor and log dSM status events");
    m_valid = true;
    m_state = kStateInit;
    m_current_temp_ts = 0;
    m_max_temp_ts = 0;
}

DsmMonitorTask::~DsmMonitorTask()
{
    try
    {
        dSA::dS_Assistant::getInstance()->dS485()->removeListener(this);
    }
    catch(Exception& e)
    {
    }
}

void DsmMonitorTask::callback(boost::shared_ptr<CBData> busdata)
{
    boost::shared_ptr<CBData> data = busdata;

    switch(busdata->getType())
    {
        case DsmApiDsmState:
        case DsmApiDsmError:
        case DsmApiDsmRebootRequest:
            break;
        default:
            return;
    }
    m_mutex.lock();
    m_queue.push(busdata);
    m_mutex.unlock();
}

void DsmMonitorTask::process()
{
    boost::shared_ptr<CBData> busdata;
    boost::shared_ptr<std::list<boost::shared_ptr<CBData> > > cbdatalist(
            new(std::list<boost::shared_ptr<CBData> >));

    m_mutex.lock();
    while (!m_queue.empty())
    {
        busdata = m_queue.front();
        m_queue.pop();

        if (busdata->getType() == DsmApiDsmState)
        {
            boost::shared_ptr<DsmStatusData> data =
                    boost::dynamic_pointer_cast<DsmStatusData> (busdata);

            DSUID_LOG(logWARNING, data->getDSMSourceDSUID()) << "[Task DsmMonitor] " <<
                    "State: Hardware:" << util::str(data->getHardware()) << ", "
                    "CommunicationDSP:" << util::str(data->getCommunicationDSP()) << ", "
                    "Temperature:" << util::str(data->getTemperature()) << ", "
                    "Config:" << util::str(data->getConfiguration()) << ", "
                    "DS485:" << util::str(data->getDS485Bus()) << ", "
                    "Line:" << util::str(data->getPowerline());

            try
            {
                Database::getInstance()->updateDSMState(data->getDSMSourceDSUID(),
                    data->getHardware(),
                    data->getCommunicationDSP(),
                    data->getTemperature(),
                    data->getConfiguration(),
                    data->getDS485Bus(),
                    data->getPowerline());
            }
            catch (Exception &ex)
            {
                FILE_LOG(logWARNING) << "[DSM Monitor Task] "
                        "Failed to store dsm status: " << ex.getMessage();
                throw ex;
            }
        } else if (busdata->getType() == DsmApiDsmError)
        {
            boost::shared_ptr<DsmErrorData> data =
                    boost::dynamic_pointer_cast<DsmErrorData> (busdata);

            DSUID_LOG(logWARNING, data->getDSMSourceDSUID()) << "[Task DsmMonitor] " <<
                    "Error: " << util::str(data->getError()) << ", "
                    "Data: " << util::str(data->getData());
        } else if (busdata->getType() == DsmApiDsmRebootRequest)
        {
            boost::shared_ptr<DsmRebootRequestData> data =
                    boost::dynamic_pointer_cast<DsmRebootRequestData> (busdata);

            DSUID_LOG(logINFO, data->getDSMDestinationDSUID()) << "[Task DsmMonitor] " <<
                    "dSM got Reboot request";
        }
    }
    m_mutex.unlock();
}

void DsmMonitorTask::checkTemperature(dS_Assistant *dsa)
{
    time_t now = time(NULL);

    bool check_temp = (difftime(now, m_current_temp_ts) >=
                                                    DSM_TEMP_LOG_INTERVAL);
    bool check_max_temp = (difftime(now, m_max_temp_ts) >=
                                                    DSM_MAX_TEMP_LOG_INTERVAL);

    if (!check_temp && !check_max_temp)
    {
        return;
    }

    if (check_temp)
    {
        m_current_temp_ts = now;
    }

    if (check_max_temp)
    {
        m_max_temp_ts = now;
    }

    boost::shared_ptr<std::list<dsuid_t> > dsms;

    try
    {
        dsms = Storage::getInstance()->getDSMList();
    }
    catch (Exception &ex)
    {
        FILE_LOG(logWARNING) << "[DSM Monitor Task] "
                "Failed to load dSMs from database: "
                << ex.getMessage();
        return;
    }

    while (!dsms->empty())
    {
        dsuid_t dsuid = dsms->front();
        dsms->pop_front();

        if (check_temp)
        {
            try
            {
                uint16_t curTemp, maxTemp;
                dsa->dS485()->getDSMtemperature(dsuid, &curTemp);
                dsa->dS485()->getDSMMaxTemperature(dsuid, &maxTemp);
                DSUID_DATA_LOG(dsuid, "temp_")
                    << curTemp << "," << maxTemp;
            }
            catch (Exception &ex)
            {
                FILE_LOG(logWARNING) << "[DSM Monitor Task] "
                        "Failed to get dSM temperatures: " << ex.getMessage();
            }
        }
        if (check_max_temp)
        {
            try
            {
                dsa->dS485()->resetDSMMaxTemperature(dsuid);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logWARNING) << "[DSM Monitor Task] "
                        "Failed to reset dSM max temperature: "
                        << ex.getMessage();
            }
        }
    }
}

void DsmMonitorTask::run(dS_Assistant *dsa)
{
    // initial setup
    if (kStateInit == m_state)
    {
        // add task to receive callbacks
        dsa->dS485()->addListener(this);

        // select the callbacks you want
        this->select(DsmApiDsmState, true);
        this->select(DsmApiDsmError, true);
        this->select(DsmApiDsmRebootRequest, true);

        // next state
        m_state = kStateSetup;
    }

    while (m_valid)
    {
        switch (m_state)
        {
            case kStateSetup:
            {
                process();
                checkTemperature(dsa);
                m_state = kStateReady;
            }
            break;

            case kStateReady:
            default:
            {
                m_state = kStateSetup;
                dsa->scheduleTask(boost::shared_ptr<dSTask> (shared_from_this()), 5);
                return;
            }
            break;
        }
    }

    // final cleanup
    dsa->dS485()->removeListener(this);
}

DeviceMonitorTask::DeviceMonitorTask() : dSLongTask(),
        m_keyType(0), m_lastQuery(0)
{
    m_name = _("Device traffic monitor");
    m_description = _("Permanently monitor upstream traffic");
    m_valid = true;
    m_state = kStateInit;
}

DeviceMonitorTask::~DeviceMonitorTask()
{
    try
    {
        dSA::dS_Assistant::getInstance()->dS485()->removeListener(this);
    }
    catch(Exception& e)
    {
    }
}

task_type DeviceMonitorTask::getTaskType() const
{
    return TestDeviceMonitor;
}

std::string DeviceMonitorTask::getLastDSUID()
{
    m_status_mutex.lock();
    m_lastQuery = time(NULL);
    m_status_mutex.unlock();

    return util::str(m_devid);
}

std::string DeviceMonitorTask::getLastDSMDSUID()
{
    m_status_mutex.lock();
    m_lastQuery = time(NULL);
    m_status_mutex.unlock();

    return util::str(m_dsmid);
}

uint8_t DeviceMonitorTask::getKeyType()
{
  m_status_mutex.lock();
  m_lastQuery = time(NULL);
  m_status_mutex.unlock();

  return m_keyType;
}

void DeviceMonitorTask::callback(boost::shared_ptr<CBData> busdata)
{
    boost::shared_ptr<const Device> dsd;

    switch(busdata->getType())
    {
        default:
            break;

        case DsmApiCommUpstreamShort:
        {
            boost::shared_ptr<CommUpstreamShortData> data =
                boost::dynamic_pointer_cast<CommUpstreamShortData> (busdata);

            if (data->getCrosstalkFlags() & 1)
            {
                return;
            }

            try
            {
                dsd = Storage::getInstance()->getDevice(
                            data->getDeviceID(),
                            data->getDSMSourceDSUID());
                m_devid = dsd->getDSUID();
                m_dsmid = data->getDSMSourceDSUID();
                m_keyType = data->getKeyType();
            }
            catch (Exception &ex)
            {
                m_devid = DSUID_NULL;
            }

        }
        break;
    }
}

void DeviceMonitorTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    // initial setup
    if (kStateInit == m_state)
    {
        m_status_mutex.lock();
        m_lastQuery = time(NULL);
        m_devid = m_dsmid = DSUID_NULL;
        m_keyType = 0;
        m_status_mutex.unlock();

        // add task to receive callbacks
        dsa->dS485()->addListener(this);

        // select the callbacks you want
        this->select(DsmApiCommUpstreamShort, true);

        m_state = kStateSetup;

        FILE_LOG(logINFO) << "[DeviceMonitor Task] starting";
    }

    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid)
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                m_state = kStateReady;
            }
            break;

            case kStateReady:
            default:
            {
                m_state = kStateSetup;
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 30);
                return;
            }
            break;
        }

        // idle timeout, stop task if no one cares for the results
        time_t to;
        m_status_mutex.lock();
        to = time(NULL) - m_lastQuery;
        m_status_mutex.unlock();
        if (to > 60)
        {
            break;
        }
    }

    // final cleanup
    dsa->dS485()->removeListener(this);

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
    }
    else
    {
        m_task_status = Finished;
        m_status_info = _("Finished");
    }
    m_status_mutex.unlock();

    FILE_LOG(logINFO) << "[DeviceMonitor Task] finished";
}

DatabaseCleanupTask::DatabaseCleanupTask(size_t max_db_filesize) :
        dSTask(), m_max_filesize(max_db_filesize)
{
    m_name = _("Periodical database cleanup");
    m_description = _("Cleanup and database maintenance");
    m_valid = true;
    m_state = kStateInit;
    m_iteration = 0;
}

void DatabaseCleanupTask::run(dS_Assistant *dsa)
{
    // initial setup
    if (kStateInit == m_state)
    {
        FILE_LOG(logINFO) << "[DatabaseCleanup] " <<
                "Starting cleanup function, current db size: " <<
                util::str(Database::getInstance()->getFileSize());
        m_iteration = 0;
        m_state = kStateSetup;
    }

    while (m_valid)
    {
        switch (m_state)
        {
            case kStateSetup:
            {
                m_iteration ++;
                Database* db = Database::getInstance();
                if (db->getFileSize() > m_max_filesize)
                {
                    switch(m_iteration)
                    {
                        case 1:
                            db->removeAreaConfigurationData(DB_CLEANUP_DAYS(1));
                            break;
                        case 2:
                            db->removeUpstreamMetricByTime(DB_CLEANUP_DAYS(21));
                            break;
                        case 3:
                            db->removeUpstreamMetricByTime(DB_CLEANUP_DAYS(7));
                            break;
                        case 4:
                            db->removeUpstreamMetricByTime(DB_CLEANUP_DAYS(3));
                            break;
                        default:
                            m_state = kStateReady;
                    }
                    db->executeVacuum();
                }
                if (db->getFileSize() < m_max_filesize)
                {
                    m_state = kStateReady;
                }
            }
            break;

            case kStateReady:
            default:
            {
                // calculate next execution time, 7 days ahead at 4am
                time_t now = time(NULL);
                struct tm tmp;
                localtime_r(&now, &tmp);
                tmp.tm_mday += 7;
                tmp.tm_hour = 4;
                tmp.tm_min = 0;

                size_t dbsize = Database::getInstance()->getFileSize();
                if (dbsize > m_max_filesize)
                {
                    FILE_LOG(logWARNING) << "[DatabaseCleanup] " <<
                            "file size still too large: " <<
                            util::str(dbsize);
                    localtime_r(&now, &tmp);
                    tmp.tm_mday ++;
                    tmp.tm_hour = 4;
                    tmp.tm_min = 0;
                }
                time_t scheduledTime = mktime(&tmp);

                FILE_LOG(logINFO) << "[DatabaseCleanup] " <<
                        "schedule next maintenance in " <<
                        util::str((scheduledTime - now) / 60) << " minutes.";

                boost::shared_ptr<DatabaseCleanupTask>
                                task(new DatabaseCleanupTask(m_max_filesize));
                if (scheduledTime > now)
                {
                    dsa->scheduleTask(task, scheduledTime - now);
                }
                else
                {
                    dsa->scheduleTask(task, 12 * 60 * 60);
                }
                return;
            }
            break;
        }
    }

    // final cleanup
}

DeviceUpdateTask::DeviceUpdateTask(dsuid_t dsm_dsuid, dsuid_t dev_dsuid) :
        dSPersistentTask(), m_rvalue(0), m_configid_new(0), m_errorcount(0),
        m_dev_dsuid(dev_dsuid), m_dsm_dsuid(dsm_dsuid), m_dev_max(0)
{
    m_name = _("Device Update Task");
    m_description = _("Update device configuration data tables");
    m_parameters = "dsm=" + util::str(dsm_dsuid) + "&dsuid=" + util::str(dev_dsuid);
    m_valid = true;
    m_pausable = true;
    m_state = kStateInit;
    m_dev_count = m_errorcode = m_retry = 0;
}

DeviceUpdateTask::~DeviceUpdateTask()
{
    try
    {
        dSA::dS_Assistant::getInstance()->dS485()->removeListener(this);
    }
    catch(Exception& e)
    {
    }
}

task_type DeviceUpdateTask::getTaskType() const
{
    return TestDeviceUpdateTask;
}

std::string DeviceUpdateTask::getCurrentDSUID()
{
    return util::str(m_device->getDSUID());
}

bool DeviceUpdateTask::updateable(boost::shared_ptr<Device> dev)
{
    if (dev->getDeviceType() == dt_VirtualDevice)
    {
        return (!dev->isMultiInputDevice());
    }
    else
    {
        // beta releases
        if (dev->getVersion() >= 0xa000)
        {
            return false;
        }

        // tkm branch 3.4x
        if (((dev->getProductId() >> 10) == 1) &&
            (dev->getVersion() >= 0x0340) &&
            (dev->getVersion() <= 0x034f))
        {
            return false;
        }

        return (!dev->isMultiInputDevice());
    }
}

void DeviceUpdateTask::filterDeviceList()
{
    std::list<boost::shared_ptr<Device> >::iterator it;

    m_mutex.lock();
    for (it = m_dev_list->begin(); it != m_dev_list->end(); it ++)
    {
        if (! updateable((*it)))
        {
            it = m_dev_list->erase(it);
        }
    }
    m_mutex.unlock();
}

void DeviceUpdateTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {
        default:
            break;

        case DsmApiDeviceActive:
        {
            boost::shared_ptr<DeviceActiveData> data =
                boost::dynamic_pointer_cast<DeviceActiveData> (busdata);

            m_mutex.lock();
            m_dev_list_added.push_back(std::make_pair<dsuid_t, uint16_t>
                (data->getDSMSourceDSUID(), data->getDeviceID()));
            m_mutex.unlock();
        }
        break;
    }
}

uint8_t DeviceUpdateTask::readDeviceConfigurationId(dS_Assistant* dsa, boost::shared_ptr<Device> dev)
{
    FILE_LOG(logDEBUG) << "[DeviceUpdateTask] Reading config-id of device " <<
            util::str(dev->getDSUID());

    int retry = 0;
    int errorcode = 0;
    uint8_t value;

    do
    {
        try
        {
            if (dev->getVersion() >= 0x358)
            {
                dsa->dS485()->getConfigurationValue(dev->getDSMdSUID(), dev->getId(),
                        kConfigRevisionClass358, kConfigRevisionOffset358, &value);
            } else
            {
                dsa->dS485()->getConfigurationValue(dev->getDSMdSUID(), dev->getId(),
                        kConfigRevisionClass, kConfigRevisionOffset, &value);
            }

            FILE_LOG(logINFO) << "[DeviceUpdateTask] Device "
                    << util::str(dev->getDSUID()) << " has config-id "
                    << util::str(value);
            return value;
        }
        catch (Exception &ex)
        {
            errorcode = ex.getErrorCode();
            if (errorcode < 0 || retry >= 3)
            {
                FILE_LOG(logWARNING) << "[DeviceUpdateTask] "
                        "Failed to get configuration version from device " <<
                        util::str(dev->getDSUID()) <<
                        " [Error " << util::str(errorcode) << "]";
                throw ex;
            }
        }
    } while (retry++ < 3);

    return 255;
}

void DeviceUpdateTask::writeDeviceConfigurationId(dS_Assistant *dsa, boost::shared_ptr<Device> dev, uint8_t value)
{
    int retry = 0;
    int errorcode = 0;

    do
    {
        try
        {
            if (dev->getVersion() >= 0x358)
            {
                dsa->dS485()->setConfigurationValue(m_device->getDSMdSUID(),
                        m_device->getId(), kConfigRevisionClass358, kConfigRevisionOffset358, value);
            } else
            {
                dsa->dS485()->setConfigurationValue(m_device->getDSMdSUID(),
                        m_device->getId(), kConfigRevisionClass, kConfigRevisionOffset, value);
            }
            return;
        }
        catch (Exception &ex)
        {
            errorcode = ex.getErrorCode();
            if (errorcode < 0 || retry >= 3)
            {
                FILE_LOG(logWARNING) << "[DeviceUpdateTask] "
                        "Failed to set configuration version for device " <<
                        util::str(dev->getDSUID()) <<
                        " [Error " << util::str(errorcode) << "]";
                throw ex;
            }
        }
    } while (retry++ < 3);
}

uint8_t DeviceUpdateTask::readDeviceParameter(dS_Assistant *dsa, boost::shared_ptr<Device> dev, uint8_t bank, uint8_t offset)
{
    int retry = 0;
    int errorcode = 0;
    uint8_t rvalue;

    do
    {
        try
        {
            dsa->dS485()->getConfigurationValue(dev->getDSMdSUID(),
                    dev->getId(), bank, offset, &rvalue);
            return rvalue;
        }
        catch (Exception &ex)
        {
            errorcode = ex.getErrorCode();
            if (errorcode < 0 || retry >= 3)
            {
                FILE_LOG(logWARNING) << "[DeviceUpdateTask] "
                        "Reading device " << util::str(dev->getDSUID()) <<
                        ": Class=" << util::str(bank) <<
                        ", Offset=" << util::str(offset) <<
                        ": Failed [Error " << util::str(errorcode) << "]";
                throw ex;
            }
        }
    } while (retry++ < 3);

    return 0;
}

void DeviceUpdateTask::writeDeviceParameter(dS_Assistant *dsa, boost::shared_ptr<Device> dev,
            uint8_t bank, uint8_t offset, uint8_t value)
{
    int retry = 0;
    int errorcode = 0;

    do
    {
        try
        {
            dsa->dS485()->setConfigurationValue(m_device->getDSMdSUID(),
                    m_device->getId(), bank, offset, value);
            return;
        }
        catch (Exception &ex)
        {
            errorcode = ex.getErrorCode();
            if (errorcode < 0 || retry >= 3)
            {
                FILE_LOG(logWARNING) << "[DeviceUpdateTask] "
                        "Failed to update configuration for device " <<
                        util::str(dev->getDSUID()) <<
                        " [Error " << util::str(errorcode) << "]";
                throw ex;
            }
        }
    } while (retry++ < 3);
}

void DeviceUpdateTask::saveDeviceConfigUpdate(boost::shared_ptr<Device> device,
            uint8_t status, uint8_t errorcode)
{
    try
    {
        Database::getInstance()->saveDeviceConfigUpdate(device, status, errorcode);
    }
    catch (SQLite3Exception &sx)
    {
        FILE_LOG(logERROR) << "[DeviceUpdateTask] Error storing task results: " << sx.getMessage();
        if (SQLite3::isFatal(sx.getErrorCode()))
        {
            throw sx;
        }
    }
}

void DeviceUpdateTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    // initial setup
    if (kStateInit == m_state)
    {
        m_status_mutex.lock();
        try
        {
            if (dsuid_is_null(&m_dsm_dsuid))
            {
                m_dev_list = Storage::getInstance()->getDevices(false);
            }
            else if (dsuid_is_null(&m_dev_dsuid))
            {
                m_dev_list = Storage::getInstance()->getDevices(m_dsm_dsuid, false);
            }
            else
            {
                m_dev_list = boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                    (new std::list<boost::shared_ptr<Device> >);
                m_dev_list->push_back(
                        Storage::getInstance()->getDevice(m_dev_dsuid, m_dsm_dsuid));
            }
        }
        catch (SQLite3Exception &sx)
        {
            FILE_LOG(logERROR) << sx.getMessage();
            if (SQLite3::isFatal(sx.getErrorCode()))
            {
                throw sx;
            }
            m_dev_list = boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                (new std::list<boost::shared_ptr<Device> >);
        }
        catch (Exception &ex)
        {
            m_dev_list = boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                (new std::list<boost::shared_ptr<Device> >);
        }

        m_dev_count = 0;
        m_dev_max = m_dev_list->size();
        m_dev_current = m_dev_list->begin();
        m_progress = 0;
        m_status_mutex.unlock();

        // add task to receive callbacks
        dsa->dS485()->addListener(this);

        // select the callbacks you want
        this->select(DsmApiDeviceActive, true);

        m_state = kStateSetup;

        FILE_LOG(logINFO) << "[DeviceUpdateTask] starting";
    }

    if (m_paused)
    {
        m_paused = false;
    }

    while (m_valid)
    {
        if (m_paused)
        {
            m_status_mutex.lock();
            m_task_status = Sleeping;
            m_status_mutex.unlock();
            return;
        }

        switch (m_state)
        {
            case kStateSetup:
            {
                if (m_dev_current == m_dev_list->end())
                {
                    m_state = kStateStartupComplete;
                    break;
                }

                m_device = *m_dev_current;
                if (!updateable(m_device))
                {
                    saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_INCOMPATIBLE, DEVCONFIG_ERROR_OK);

                    FILE_LOG(logINFO) << "[DeviceUpdateTask] "
                            "Skip device " << util::str(m_device->getDSUID()) << ", cannot be updated";
                    m_dev_current = m_dev_list->erase(m_dev_current);
                    m_dev_count ++;
                    break;
                }
                saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_UNKNOWN, DEVCONFIG_ERROR_OK);

                m_dev_count ++;
                m_status_mutex.lock();
                m_progress = (unsigned char) ((m_dev_count-1) * 100 / m_dev_max);
                m_progress_info = _P(_("Analyzing configuration: device %d of %d"),
                                   m_dev_count, m_dev_max);
                m_status_mutex.unlock();
                FILE_LOG(logDEBUG) << "PROGRESS: " << m_progress_info;

                m_state = kStateStartup;
            }
            break;

            case kStateStartup:
            {
                m_state = kStateStartupWait;
                if (m_device->getConfigurationId() == 255) {
                    if (m_device->getState() == 0)
                    {
                        saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_HASERRORS, DEVCONFIG_ERROR_INACTIVE);
                        FILE_LOG(logINFO) << "[DeviceUpdateTask] "
                                "Skip device " << util::str(m_device->getDSUID()) << ", is inactive";
                        m_dev_current = m_dev_list->erase(m_dev_current);
                        break;
                    }
                    try
                    {
                        uint8_t value = readDeviceConfigurationId(dsa, m_device);
                        m_device->setConfigurationId(value);
                        Database::getInstance()->saveDevice(m_device);
                    }
                    catch (SQLite3Exception &sx)
                    {
                        FILE_LOG(logERROR) << "[DeviceUpdateTask] " << sx.getMessage();
                        if (SQLite3::isFatal(sx.getErrorCode()))
                        {
                            throw sx;
                        }
                    }
                    catch (Exception& ex)
                    {
                        saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_HASERRORS, DEVCONFIG_ERROR_COMM);
                        FILE_LOG(logWARNING) << "[DeviceUpdateTask] "
                                "Skip device " << util::str(m_device->getDSUID()) << ", could not read config id";
                        m_dev_current = m_dev_list->erase(m_dev_current);
                    }

                    // do reschedule task in a few seconds to avoid bus monopolization
                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), kRescheduleTime);
                    return;
                }
            }
            break;

            case kStateStartupWait:
            {
                if (m_device->getConfigurationId() != 255)
                {
                    DeviceConfigReader reader;
                    boost::shared_ptr<DeviceConfig> configVer0 = reader.read(*m_device.get(),
                            m_device->getConfigurationId());
                    boost::shared_ptr<DeviceConfig> configVer1 = reader.read(*m_device.get(), -1);

                    if (configVer0 && configVer1)
                    {
                        std::vector<int> devSceneChanges = configVer0->sceneChanges(*configVer1);
                        if (!devSceneChanges.empty())
                        {
                            FILE_LOG(logDEBUG) << "[DeviceUpdateTask] adding " << devSceneChanges.size()
                                    << " config changes for device " << util::str(m_device->getDSUID());
                            m_scene_changes.insert(m_scene_changes.end(), devSceneChanges.begin(),
                                    devSceneChanges.end());
                            std::sort(m_scene_changes.begin(), m_scene_changes.end());
                            m_scene_changes.erase(std::unique(m_scene_changes.begin(), m_scene_changes.end()),
                                    m_scene_changes.end());
                        }
                        if (configVer1->getVersion() == m_device->getConfigurationId())
                        {
                            saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_UPTODATE, DEVCONFIG_ERROR_OK);
                            m_state = kStateReady;
                        }
                    }
                    m_dev_current++;
                }

                m_state = kStateSetup;
                if (m_dev_current == m_dev_list->end())
                {
                    m_state = kStateStartupComplete;
                }
            }
            break;

            case kStateStartupComplete:
            {
                // again iterate over all devices and actually update device configuration
                if (m_dev_list->size() > 0)
                {
                    m_dev_count = 0;
                    m_dev_max = m_dev_list->size();
                    m_dev_current = m_dev_list->begin();
                    m_state = kStateVersion;
                }
                else
                {
                    m_state = kStateFinished;
                    break;
                }

                if (m_scene_changes.empty())
                {
                    m_state = kStateFinished;
                }
                else
                {
                    FILE_LOG(logINFO) << "[DeviceUpdateTask] "
                            "Have to update " << util::str(m_scene_changes.size()) << " scenes";
                    for (std::vector<int>::iterator it = m_scene_changes.begin();
                            it != m_scene_changes.end();
                            it ++)
                    {
                        FILE_LOG(logDEBUG) << "[DeviceUpdateTask] "
                                " update scene number " << util::str(*it);
                    }

                    CommChannel::sendSceneLocks(m_scene_changes);
                }
            }
            break;

            case kStateVersion:
            {
                m_device = *m_dev_current;
                m_errorcount = 0;
                m_dev_count ++;

                m_status_mutex.lock();
                if (m_dev_max > 0)
                {
                    m_progress = (unsigned char) ((m_dev_count-1) * 100 / m_dev_max);
                }
                else
                {
                    m_progress = 100;
                }
                m_progress_info = _P(_("Updating configuration: device %d of %d"),
                                   m_dev_count, m_dev_max);
                m_status_mutex.unlock();
                FILE_LOG(logDEBUG) << "PROGRESS: " << m_progress_info;

                m_state = kStateRequest;

                DeviceConfigReader reader;
                boost::shared_ptr<DeviceConfig> configVer0 = reader.read(* m_device.get(),
                        m_device->getConfigurationId());
                boost::shared_ptr<DeviceConfig> configVer1 = reader.read(* m_device.get(),
                        -1);

                if (!configVer0 || !configVer1)
                {
                    FILE_LOG(logWARNING) << "[DeviceUpdateTask] "
                            "No configuration data found for device " <<
                            util::str(m_device->getDSUID());
                    m_state = kStateReady;
                }
                else if (configVer1->getVersion() == m_device->getConfigurationId())
                {
                    saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_UPTODATE, DEVCONFIG_ERROR_OK);
                    m_state = kStateReady;
                }
                else
                {
                    FILE_LOG(logINFO) << "[DeviceUpdateTask] "
                            "Update configuration of device " << util::str(m_device->getDSUID()) <<
                            " from version " << util::str(configVer0->getVersion()) <<
                            " to version " << util::str(configVer1->getVersion());
                    m_configid_new = configVer1->getVersion();
                    m_changes = configVer0->parameterChanges((*configVer1));
                    m_change_current = m_changes.begin();
                    if (m_change_current == m_changes.end())
                    {
                        m_state = kStateNext;
                    }
                    else {
                        saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_INPROGRESS, DEVCONFIG_ERROR_OK);
                        m_state = kStateRequest;
                    }
                }
            }
            break;

            case kStateRequest:
            {
                if (m_device->getState() == 0)
                {
                    saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_HASERRORS, DEVCONFIG_ERROR_INACTIVE);

                    FILE_LOG(logWARNING) << "[DeviceUpdateTask] "
                            "Abort updating device " << util::str(m_device->getDSUID()) << ", is inactive";
                    m_errorcount ++;
                    m_state = kStateNext;
                    m_dev_current = m_dev_list->erase(m_dev_current);
                    break;
                }

                m_status_mutex.lock();
                m_progress = (unsigned char) ((m_dev_count-1) * 100 / m_dev_max);
                m_progress_info = _P(_("Updating configuration: device %d of %d, dataset %d of %d"),
                                   m_dev_count, m_dev_max,
                                   std::distance(m_changes.begin(), m_change_current),
                                   m_changes.size());
                m_status_mutex.unlock();
                FILE_LOG(logDEBUG) << "PROGRESS: " << m_progress_info;

                m_state = kStateResponse;
                m_rvalue = 0;
                if ((*m_change_current)->mask != 0xff)
                {
                    try
                    {
                        m_rvalue = readDeviceParameter(dsa, m_device,
                                (*m_change_current)->bank, (*m_change_current)->offset);

                        FILE_LOG(logINFO) << "[DeviceUpdateTask] "
                                "Reading device " << util::str(m_device->getDSUID()) <<
                                ": Class=" << util::str((*m_change_current)->bank) <<
                                ", Offset=" << util::str((*m_change_current)->offset) <<
                                ", Value=" << util::str(m_rvalue);
                    }
                    catch (Exception& ex)
                    {
                        saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_HASERRORS, DEVCONFIG_ERROR_COMM);

                        FILE_LOG(logWARNING) << "[DeviceUpdateTask] "
                                "Abort updating device " << util::str(m_device->getDSUID()) << ", " << ex.getMessage();
                        m_errorcount ++;
                        m_state = kStateNext;
                    }

                }
                m_state = kStateResponse;
            }
            break;

            case kStateResponse:
            {
                uint8_t value;
                if ((*m_change_current)->mask != 0xff)
                {
                    value = m_rvalue & ~ (*m_change_current)->mask;
                    value |= (*m_change_current)->value;
                }
                else {
                    value = (*m_change_current)->value;
                }

                try
                {
                    FILE_LOG(logINFO) << "[DeviceUpdateTask] "
                            "Writing device " << util::str(m_device->getDSUID()) <<
                            ": Class=" << util::str((*m_change_current)->bank) <<
                            ", Offset=" << util::str((*m_change_current)->offset) <<
                            ", Mask=" << util::str((*m_change_current)->mask) <<
                            ", Default=" << util::str((*m_change_current)->value) <<
                            ", Value=" << util::str(value);

                    writeDeviceParameter(dsa, m_device,
                            (*m_change_current)->bank, (*m_change_current)->offset, value);
                    m_state = kStateNext;

                    // do reschedule task in a few seconds to avoid bus monopolization
                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), kRescheduleTime);
                    return;
                }
                catch (Exception& ex)
                {
                    saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_HASERRORS, DEVCONFIG_ERROR_COMM);
                    m_errorcount ++;
                    m_state = kStateReady;
                }
            }
            break;

            case kStateNext:
            {
                m_state = kStateRequest;

                // next configuration change
                if (m_change_current != m_changes.end())
                {
                    m_change_current++;
                }

                // all changes done
                if (m_change_current == m_changes.end())
                {
                    m_state = kStateReady;

                    // update the configuration id
                    if (m_errorcount == 0)
                    {
                        try
                        {
                            writeDeviceConfigurationId(dsa, m_device, m_configid_new);
                            m_device->setConfigurationId(m_configid_new);
                            Database::getInstance()->saveDevice(m_device);
                            saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_UPTODATE, DEVCONFIG_ERROR_OK);
                        }
                        catch (SQLite3Exception &sx)
                        {
                            FILE_LOG(logERROR) << "[DeviceUpdateTask] " << sx.getMessage();
                            if (SQLite3::isFatal(sx.getErrorCode()))
                            {
                                throw sx;
                            }
                        }
                        catch (Exception &ex)
                        {
                            saveDeviceConfigUpdate(m_device, DEVCONFIG_STATUS_HASERRORS, DEVCONFIG_ERROR_COMM);
                        }
                    }
                }

                // do reschedule task in a few seconds to avoid bus monopolization
                dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), kRescheduleTime);
                return;
            }
            break;

            case kStateReady:
            {
                m_state = kStateVersion;
                m_dev_current++;
                if (m_dev_current == m_dev_list->end())
                {
                    m_state = kStateFinished;
                }
            }
            break;

            case kStateFinished:
            default:
            {
                std::vector<int> empty;
                CommChannel::sendSceneLocks(empty);
                m_scene_changes.clear();

                m_status_mutex.lock();
                m_progress = 100;
                m_progress_info = _("Update completed");
                m_status_mutex.unlock();
                FILE_LOG(logDEBUG) << "PROGRESS: " << m_progress_info;

                FILE_LOG(logINFO) << "[DeviceUpdateTask] finished";
                m_state = kStateWait;
            }
            break;

            case kStateWait:
            {
                m_dev_count = 0;
                m_dev_list->clear();

                m_mutex.lock();
                if (m_dev_list_added.size())
                {
                    for (std::vector<std::pair<dsuid_t, uint16_t> >::iterator it = m_dev_list_added.begin();
                            it != m_dev_list_added.end();
                            )
                    {
                        try
                        {
                            boost::shared_ptr<Device> dsd =
                                    Storage::getInstance()->getDevice(it->second, it->first);
                            if (dsd)
                            {
                                m_dev_list->push_back(dsd);
                                it = m_dev_list_added.erase(it);
                            }
                            else
                            {
                                it ++;
                            }
                        }
                        catch (Exception &ex)
                        {
                            // #2491: endless loop if device has been removed in the mean time
                            // erase added device in the list and do not keep it
                            it = m_dev_list_added.erase(it);
                        }
                    }
                    m_mutex.unlock();

                    m_dev_max = m_dev_list->size();
                    m_dev_current = m_dev_list->begin();
                    m_progress = 0;
                    m_state = kStateSetup;
                    FILE_LOG(logINFO) << "[DeviceUpdateTask] restarting";
                }
                else
                {
                    m_mutex.unlock();

                    // this tasks never finishes - reschedule task execution to check for added devices
                    m_state = kStateWait;
                    dsa->scheduleLongTask(boost::shared_ptr<dSLongTask> (shared_from_this()), 30);
                    return;
                }
            }
            break;

        }
    }

    // final cleanup
    dsa->dS485()->removeListener(this);

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
    }
    else
    {
        m_task_status = Finished;
        m_status_info = _("Finished");
    }
    m_status_mutex.unlock();
}

ModifyDevicePulseTask::ModifyDevicePulseTask(int schedule) : dSPersistentTask()
{
    m_name = _("Modify device pulse settings");
    m_description = _("Iterate through all devices and modify their pulse "
                      "settings");
    m_valid = true;
    m_initialized = false;
    m_temperature_read = false;
    m_schedule = schedule;
}

task_type ModifyDevicePulseTask::getTaskType() const
{
    return ModifyDevicePulse;
}

void ModifyDevicePulseTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        boost::mutex::scoped_lock lock(m_status_mutex);
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
        return;
    }

    if (m_finished)
    {
        boost::mutex::scoped_lock lock(m_status_mutex);
        m_task_status = Finished;
        m_status_info = _("Finished");
        return;
    }

    if (m_schedule > 0)
    {
        int temp = m_schedule;
        m_schedule = 0;
        m_initialized = false;
        m_temperature_read = false;
        dsa->scheduleLongTask(
            boost::shared_ptr<dSPersistentTask> (shared_from_this()), temp);
        return;
    }

    try
    {
        if (!m_initialized && !m_temperature_read)
        {
            // get all active devices
            m_dev_list = Storage::getInstance()->getDevices(true);
            m_initialized = true;
            FILE_LOG(logINFO) << "[Task ModifyDevicesPulse] start reading temperature values";
        }

        boost::shared_ptr<Device> device;

        if (!m_temperature_read)
        {
            while (!m_dev_list->empty())
            {
                device = m_dev_list->front();
                m_dev_list->pop_front();

                // ignore multi-dSUID slaves, SDS-220 and TNY
                if (!device->isMainDevice() ||
                    (device->getDeviceType() == dt_VirtualDevice) ||
                    (device->getHardwareInfoString() == "GE-SDS200") ||
                    (device->getHardwareInfoString() == "GE-SDS220") ||
                    (device->getHardwareInfoString() == "SW-TNY200") ||
                    (device->getHardwareInfoString() == "GE-UMV200") ||
                    (device->getHardwareInfoString() == "SW-SK200"))
                {
                    FILE_LOG(logDEBUG1) << "[Task ModifyDevicesPulse] "
                            "do not read temperature from device " <<
                            device->getHardwareInfoString() << " : " <<
                            util::str(device->getDSUID());
                    device.reset();
                    continue;
                }

                break;
            }

            if (device == NULL)
            {
                m_temperature_read = true;
                m_dsms = Storage::getInstance()->getDSMList();
            }
            else
            {
                try
                {
                    {
                        boost::mutex::scoped_lock lock(m_status_mutex);
                        m_progress_info = _("Querying temperature from devices");
                    }

                    uint16_t val;
                    dsa->dS485()->getSensorValue(device->getDSMdSUID(),
                                                 device->getId(), 1, &val);
                    device->setTemperature(attributeToTemperature(val));
                    Database::getInstance()->saveDevice(device);
                }
                catch (Exception &ex2)
                {
                    FILE_LOG(logERROR) << "[Task ModifyDevicesPulse] Error "
                        "reading sensor values: " << ex2.getMessage();
                }
            }
        }
        else
        {
            // at this point the device information is up to date
            while (m_current_assignments.empty())
            {
                if (m_dsms->empty())
                {
                    break;
                }

                dsuid_t dsm_dsuid = m_dsms->front();
                m_dsms->pop_front();

                try
                {

                    boost::shared_ptr<dSM> dsm = Storage::getInstance()->getDSM(
                        dsm_dsuid);
                    if (dsm->isVirtual())
                    {
                        continue;
                    }
                    boost::shared_ptr<
                        std::list<boost::shared_ptr<Device> > > devices =
                            Storage::getInstance()->getDevices(dsm_dsuid, true);

                    std::list<boost::shared_ptr<Device> >::iterator it =
                        devices->begin();
                    while (it != devices->end())
                    {
                        boost::shared_ptr<Device> dev = *it;
                        if (dev != NULL)
                        {
                            if (!dev->isMainDevice() ||
                                (dev->getDeviceType() == dt_VirtualDevice) ||
                                (dev->getHardwareInfoString() == "GE-SDS200") ||
                                (dev->getHardwareInfoString() == "GE-SDS220") ||
                                (dev->getHardwareInfoString() == "SW-TNY200") ||
                                (dev->getHardwareInfoString() == "GE-UMV200") ||
                                (dev->getHardwareInfoString() == "SW-SK200"))
                            {
                                FILE_LOG(logDEBUG1) << "[Task ModifyDevicesPulse] skip device " <<
                                        dev->getHardwareInfoString() << " : " <<
                                        util::str(dev->getDSUID());
                                it = devices->erase(it);
                                continue;
                            }
                        }
                        it++;
                    }

                    FILE_LOG(logINFO) << "[Task ModifyDevicesPulse] assigning new timings "
                            " on dSM " << util::str(dsm_dsuid);

                    m_current_assignments = assignChargingPulse(*devices);
                }
                catch (Exception &devperdsm)
                {
                    FILE_LOG(logERROR) << "[Task ModifyDevicesPulse] Error "
                        "getting devices: " << devperdsm.getMessage();
                }
            }

            if (m_current_assignments.empty())
            {
                if (m_dsms->empty())
                {
                    FILE_LOG(logINFO) << "[Task ModifyDevicesPulse] finished";

                    // reque to start over
                    {
                        boost::mutex::scoped_lock lock(m_status_mutex);
                        m_progress_info = "";
                    }

                    m_temperature_read = false;
                    m_initialized = false;

                    ConfigManager::getInstance()->setParameter(
                        "user.digitalstrom.tasks.device-pulse.last-run",
                        util::str(time(NULL)));

                    dsa->scheduleLongTask(
                        boost::shared_ptr<dSPersistentTask> (shared_from_this()),
                        PULSE_MODIFICATION_TASK_INTERVAL);
                    return;
                }
            }
            else
            {
                cp_assignment_t assignment = m_current_assignments.back();
                m_current_assignments.pop_back();
                if (assignment.first->isMainDevice() &&
                    (assignment.first->getDeviceType() != dt_VirtualDevice))
                {
                    try
                    {
                        setChargingPulse(*assignment.first, assignment.second);
                    }
                    catch (Exception &scpex)
                    {
                        FILE_LOG(logERROR) <<
                            "[Task ModifyDevicesPulse] Error, could not set " <<
                            "charging pulse on device: " << scpex.getMessage();
                    }
                }
            }
        }
        // reque task to continue with the next device
        dsa->scheduleLongTask(
                boost::shared_ptr<dSPersistentTask> (shared_from_this()), 2);
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "[Task ModifyDevicesPulse] Error: "
            << ex.getMessage();
    }
}

void ModifyDevicePulseTask::invalidate(bool shutdown)
{
    dSPersistentTask::invalidate(shutdown);

    if (!shutdown)
    {
        ConfigManager::getInstance()->setParameter(
            "user.digitalstrom.tasks.device-pulse.active", false);
    }
}

ResetDevicePulseTask::ResetDevicePulseTask() : dSLongTask()
{
    m_name = _("Reset device pulse settings");
    m_description = _("Iterate through all devices and apply default pulse "
                      "settings");
    m_valid = true;
    m_initialized = false;
}

void ResetDevicePulseTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        boost::mutex::scoped_lock lock(m_status_mutex);
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
        return;
    }

    if (m_finished)
    {
        boost::mutex::scoped_lock lock(m_status_mutex);
        m_task_status = Finished;
        m_status_info = _("Finished");
        return;
    }

    try
    {
        if (!m_initialized)
        {
            // get all active devices
            m_dev_list =
                Storage::getInstance()->getDevices(true);

            m_initialized = true;
            FILE_LOG(logINFO) << "[Task ResetDevicesPulse] start resetting timings";
        }

        if (m_initialized)
        {
            boost::shared_ptr<Device> device;

            while (!m_dev_list->empty())
            {
                device = m_dev_list->front();
                m_dev_list->pop_front();

                // ignore multi-dSUID slaves
                if (!device->isMainDevice() ||
                    (device->getDeviceType() == dt_VirtualDevice) ||
                    (device->getHardwareInfoString() == "GE-SDS200") ||
                    (device->getHardwareInfoString() == "GE-SDS220") ||
                    (device->getHardwareInfoString() == "SW-TNY200") ||
                    (device->getHardwareInfoString() == "GE-UMV200") ||
                    (device->getHardwareInfoString() == "SW-SK200"))
                {
                    FILE_LOG(logDEBUG1) << "[Task ResetDevicesPulse] skip device " <<
                            device->getHardwareInfoString() << " : " <<
                            util::str(device->getDSUID());
                    device.reset();
                    continue;
                }

                break;
            }

            if (device == NULL)
            {
                boost::mutex::scoped_lock lock(m_status_mutex);
                m_task_status = Finished;
                m_status_info = _("Finished");
                FILE_LOG(logINFO) << "[Task ResetDevicesPulse] finished";
                return;
            }
            else
            {
                try
                {
                    resetChargingPulse(device);
                }
                catch (Exception &ex2)
                {
                    FILE_LOG(logERROR) << "[Task ResetDevicesPulse] Error "
                        "resetting device pulse " << ex2.getMessage();
                }
            }
        }
        // reque task to continue with the next device
        dsa->scheduleLongTask(
                boost::shared_ptr<dSLongTask> (shared_from_this()), 2);
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "[Task ResetDevicesPulse] Error: "
            << ex.getMessage();
    }
}

task_type ResetDevicePulseTask::getTaskType() const
{
    return ResetDevicePulse;
}

ReadAreaConfigurationTask::ReadAreaConfigurationTask(uint16_t zone,
                                                     uint16_t group)
                                                                : dSLongTask()
{
    m_name = _("Read area configuration of devices");
    m_description = _("Iterate through all devices of a given zone and group "
                      "and read their area configuration settings");
    m_valid = true;
    m_dev_count = 0;
    m_initialized = false;
    m_zone = zone;
    m_group = group;
    m_state = kStateInit;
    m_dev_list = boost::make_shared<std::list<boost::shared_ptr<Device> > >();
}

const int ReadAreaConfigurationTask::CfgClassScene = 0x80;

// the following function was borrowed from the dSS project
bool ReadAreaConfigurationTask::getDontCareFlag(dS_Assistant *dsa,
                                                boost::shared_ptr<Device> device,
                                                uint8_t scene)
{
    uint8_t value;
    dsa->dS485()->getConfigurationValue(device->getDSMdSUID(), device->getId(),
                          CfgClassScene, 128 + scene, &value);
    return ((value & 1) > 0);
}

bool ReadAreaConfigurationTask::isInArea(dS_Assistant *dsa,
                                         boost::shared_ptr<Device> device,
                                         uint8_t area)
{
    bool dc = getDontCareFlag(dsa, device, area);
    if (!dc)
    {
        return true;
    }
    else
    {
        return false;
    }
}

void ReadAreaConfigurationTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        boost::mutex::scoped_lock lock(m_status_mutex);
        m_task_status = Cancelled;
        m_status_info = _("Cancelled");
        return;
    }

    if (m_finished)
    {
        boost::mutex::scoped_lock lock(m_status_mutex);
        m_task_status = Finished;
        m_status_info = _("Finished");
        m_progress = 100;
        return;
    }

    {
        boost::mutex::scoped_lock lock(m_status_mutex);
        m_task_status = Running;
        m_status_info = _("Running");
    }

    try
    {
        if (!m_initialized)
        {
            // get all active devices
            boost::shared_ptr<std::list<boost::shared_ptr<Device> > > dev_list = Storage::getInstance()->getDeviceList(m_zone, true);

            std::list<boost::shared_ptr<Device> >::iterator iter = dev_list->begin();
            for (; iter != dev_list->end(); ++iter)
            {
                if ((*iter)->getOutputMode() == 0)
                {
                    continue;
                }
                if ((*iter)->getStandardColor() == 8)
                {
                    //TODO: check if device is in m_group
                    //      for now read areas of all joker devices as we don't know the groups
                }
                else if ((*iter)->getStandardColor() != m_group)
                {
                    continue;
                }
                m_dev_list->push_back(*iter);
            }

            m_dev_count = m_dev_list->size();
            m_initialized = true;
            FILE_LOG(logINFO) <<
                "[Task ReadAreaConfiguration] start reading out devices";
        }

        if (m_initialized)
        {
            boost::shared_ptr<Device> device = m_dev_list->front();
            m_dev_list->pop_front();

            FILE_LOG(logINFO) << "[Task ReadAreaConfiguration] will try device " << util::str(device->getDSUID());
            try
            {
                bool area1 = isInArea(dsa, device, 6);
                bool area2 = isInArea(dsa, device, 7);
                bool area3 = isInArea(dsa, device, 8);
                bool area4 = isInArea(dsa, device, 9);


                FILE_LOG(logINFO) << "[Task ReadAreaConfiguration] readout compete, saving...";
                Database::getInstance()->saveAreaConfigurationData(
                        m_task_id, util::str(device->getDSUID()),
                        m_zone, m_group, area1, area2, area3, area4);
            }
            catch (Exception &ex2)
            {
                FILE_LOG(logERROR) << "[Task ReadAreaConfiguration] Error "
                    "reading device area configuration " <<
                    ex2.getMessage();
            }

            if (m_dev_list->empty())
            {
                boost::mutex::scoped_lock lock(m_status_mutex);
                m_task_status = Finished;
                m_status_info = _("Finished");
                m_progress = 100;
                FILE_LOG(logINFO) << "[Task ReadAreaConfiguration] finished";
                return;
            }

            boost::mutex::scoped_lock lock(m_status_mutex);
            if (m_dev_count > 0)
            {
                m_progress = 100 * (m_dev_count - m_dev_list->size()) /
                             m_dev_count;
            }
        }
        // requeue task to continue with the next device
        dsa->scheduleLongTask(
                boost::shared_ptr<dSLongTask> (shared_from_this()), 2);
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "[Task ReadAreaConfiguration] Error: "
            << ex.getMessage();
    }
}

task_type ReadAreaConfigurationTask::getTaskType() const
{
    return ReadAreaConfiguration;
}

} // namespace
